from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, List, Optional
from datetime import timedelta

import httpx
from pydantic import Field, BaseModel, ConfigDict

if TYPE_CHECKING:
    from .api import ResultsResponse
    from .._client import Stratix, AsyncStratix


class EvaluationStatus(str, Enum):
    PENDING = "pending"
    FAILURE = "failure"
    IN_PROGRESS = "in-progress"
    PAUSED = "paused"
    SUCCESS = "success"
    TIMEOUT = "timeout"
    CANCELLED = "cancelled"


class EvaluationMetric(BaseModel):
    name: str
    description: str = ""


class EvaluationTaskType(BaseModel):
    name: str
    description: str = ""


class EvaluationDataset(BaseModel):
    total_size: int = 0
    training_size: int = 0
    test_size: int = 0
    characteristics: Optional[List[str]] = []


class EvaluationModelInfo(BaseModel):
    model_config = ConfigDict(protected_namespaces=())

    model_name: str = ""
    performance: Any = None


class PerformanceDetails(BaseModel):
    strengths: Optional[List[str]] = []
    challenges: Optional[List[str]] = []


class ErrorAnalysis(BaseModel):
    common_failure_modes: Optional[List[str]] = []
    example: str = ""


class AnalysisSummary(BaseModel):
    key_takeaways: Optional[List[str]] = []


class EvaluationSummary(BaseModel):
    name: str = ""
    goal: str = ""
    metrics: Optional[List[EvaluationMetric]] = []
    task_types: Optional[List[EvaluationTaskType]] = []
    dataset: Optional[EvaluationDataset] = None
    model: Optional[EvaluationModelInfo] = None
    performance_details: Optional[PerformanceDetails] = None
    error_analysis: Optional[ErrorAnalysis] = None
    analysis_summary: Optional[AnalysisSummary] = None


class Evaluation(BaseModel):
    model_config = ConfigDict(populate_by_name=True, protected_namespaces=())

    id: str
    status: EvaluationStatus
    status_description: str = ""
    submitted_at: int
    finished_at: int
    model_id: str
    model_name: str = ""
    model_key: str = ""
    model_company: str = ""
    benchmark_id: str = Field(..., alias="dataset_id")
    benchmark_name: str = Field("", alias="dataset_name")
    average_duration: int
    accuracy: float
    readability_score: float = 0.0
    toxicity_score: float = 0.0
    ethics_score: float = 0.0
    failed_prompt_count: int = 0
    queue_id: int = 0
    summary: Optional[EvaluationSummary] = None

    _client: "Optional[Stratix | AsyncStratix]" = None

    def attach_client(self, client: "Stratix | AsyncStratix") -> "Evaluation":
        self._client = client
        return self

    @property
    def is_finished(self) -> bool:
        """Return True if evaluation is done (success, failure, or timeout)."""
        return self.status in {
            EvaluationStatus.SUCCESS,
            EvaluationStatus.FAILURE,
            EvaluationStatus.TIMEOUT,
        }

    @property
    def is_success(self) -> bool:
        """Return True if evaluation completed successfully."""
        return self.status == EvaluationStatus.SUCCESS

    def get_results(
        self,
        *,
        page: Optional[int] = None,
        page_size: Optional[int] = None,
        timeout: float | httpx.Timeout | None = None,
    ) -> Optional[ResultsResponse]:
        """Fetch results synchronously if a sync client is attached."""
        from .._client import AsyncStratix

        if self._client is None:
            raise ValueError("No client attached")
        if isinstance(self._client, AsyncStratix):
            raise RuntimeError("Use `await get_results_async()` with an async client")

        return self._client.results.get(evaluation=self, page=page, page_size=page_size, timeout=timeout)

    def get_all_results(
        self,
        *,
        timeout: float | httpx.Timeout | None = None,
    ) -> List[Result]:
        """Fetch results synchronously if a sync client is attached."""
        from .._client import AsyncStratix

        if self._client is None:
            raise ValueError("No client attached")
        if isinstance(self._client, AsyncStratix):
            raise RuntimeError("Use `await get_results_async()` with an async client")

        return self._client.results.get_all(evaluation=self, timeout=timeout)

    async def get_results_async(
        self,
        *,
        page: Optional[int] = None,
        page_size: Optional[int] = None,
        timeout: float | httpx.Timeout | None = None,
    ) -> Optional[ResultsResponse]:
        """Fetch results asynchronously if an async client is attached."""
        from .._client import AsyncStratix

        if self._client is None:
            raise ValueError("No client attached")
        if not isinstance(self._client, AsyncStratix):
            raise RuntimeError("Use `get_results()` with a sync client")

        return await self._client.results.get(evaluation=self, page=page, page_size=page_size, timeout=timeout)

    async def get_all_results_async(
        self,
        *,
        timeout: float | httpx.Timeout | None = None,
    ) -> List[Result]:
        """Fetch results asynchronously if an async client is attached."""
        from .._client import AsyncStratix

        if self._client is None:
            raise ValueError("No client attached")
        if not isinstance(self._client, AsyncStratix):
            raise RuntimeError("Use `get_results()` with a sync client")

        return await self._client.results.get_all(evaluation=self, timeout=timeout)

    def wait_for_completion(
        self, *, interval_seconds: int = 30, timeout_seconds: Optional[int] = None
    ) -> Optional["Evaluation"]:
        """Sync polling using a sync client."""
        from .._client import AsyncStratix

        if self._client is None:
            raise ValueError("No client attached")
        if isinstance(self._client, AsyncStratix):
            raise RuntimeError("Use `wait_for_completion_async()` with an async client")

        evaluation = self._client.evaluations.wait_for_completion(
            self, interval_seconds=interval_seconds, timeout_seconds=timeout_seconds
        )
        if evaluation:
            self.status = evaluation.status
            self.status_description = evaluation.status_description
            self.finished_at = evaluation.finished_at
            self.average_duration = evaluation.average_duration
            self.accuracy = evaluation.accuracy
            self.readability_score = evaluation.readability_score
            self.toxicity_score = evaluation.toxicity_score
            self.ethics_score = evaluation.ethics_score
            self.failed_prompt_count = evaluation.failed_prompt_count
            self.summary = evaluation.summary

        return self

    async def wait_for_completion_async(
        self, *, interval_seconds: int = 30, timeout_seconds: Optional[int] = None
    ) -> Optional["Evaluation"]:
        """Async polling using an async client."""
        from .._client import AsyncStratix

        if self._client is None:
            raise ValueError("No client attached")
        if not isinstance(self._client, AsyncStratix):
            raise RuntimeError("Use `wait_for_completion()` with a sync client")

        evaluation = await self._client.evaluations.wait_for_completion(
            self, interval_seconds=interval_seconds, timeout_seconds=timeout_seconds
        )
        if evaluation:
            self.status = evaluation.status
            self.status_description = evaluation.status_description
            self.finished_at = evaluation.finished_at
            self.average_duration = evaluation.average_duration
            self.accuracy = evaluation.accuracy
            self.readability_score = evaluation.readability_score
            self.toxicity_score = evaluation.toxicity_score
            self.ethics_score = evaluation.ethics_score
            self.failed_prompt_count = evaluation.failed_prompt_count
            self.summary = evaluation.summary

        return self


class Result(BaseModel):
    subset: str
    prompt: str
    result: str
    truth: str
    duration: timedelta
    score: float
    metrics: Dict[str, Optional[float]]



from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel


class EvaluationSpaceModelFilter(BaseModel):
    ids: List[str] = []
    vendors: List[str] = []
    regions: List[str] = []


class EvaluationSpaceDatasetFilter(BaseModel):
    ids: List[str] = []
    categories: List[str] = []
    languages: List[str] = []


class EvaluationSpaceFilters(BaseModel):
    model_filters: Optional[EvaluationSpaceModelFilter] = None
    dataset_filters: Optional[EvaluationSpaceDatasetFilter] = None
    providers: List[str] = []


class EvaluationSpace(BaseModel):
    id: str
    organization_id: Optional[str] = None
    project_id: Optional[str] = None
    name: str
    description: Optional[str] = None
    filters: Optional[EvaluationSpaceFilters] = None
    owner: Optional[str] = None
    visibility: Optional[str] = None
    is_featured: bool = False
    is_partner: bool = False
    partner_name: Optional[str] = None
    created_at: Optional[str] = None
    image_path: Optional[str] = None
    weight: int = 0
    slug: Optional[str] = None
    models_count: int = 0
    benchmarks_count: int = 0
    evaluations_count: int = 0



from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel, ConfigDict


class JudgeVersion(BaseModel):
    model_config = ConfigDict(protected_namespaces=())

    version: int
    name: str
    evaluation_goal: str
    model_id: Optional[str] = None
    model_name: Optional[str] = None
    model_company: Optional[str] = None
    updated_at: str
    updated_by: str


class Judge(BaseModel):
    model_config = ConfigDict(protected_namespaces=())

    id: str
    organization_id: str
    project_id: str
    name: str
    evaluation_goal: str
    model_id: Optional[str] = None
    model_name: Optional[str] = None
    model_company: Optional[str] = None
    version: int
    run_count: int = 0
    created_at: str
    updated_at: str
    versions: List[JudgeVersion] = []



from datetime import timedelta
from unittest.mock import Mock, patch

import httpx
import pytest

from layerlens import Stratix
from layerlens.models import (
    Model,
    Result,
    Benchmark,
    Evaluation,
    EvaluationStatus,
    CreateEvaluationsResponse,
)


class TestStratixIntegration:
    """Integration tests for full Stratix API workflows."""

    @pytest.fixture
    def stratix_client(self):
        """Create Stratix client with mocked dependencies."""
        return Stratix(api_key="test-api-key")

    @pytest.fixture
    def sample_model_data(self):
        """Sample model data for testing."""
        return {
            "id": "model-gpt4",
            "key": "gpt-4",
            "name": "GPT-4",
            "company": "OpenAI",
            "description": "Large language model",
            "released_at": 1679875200,
            "parameters": 1.76e12,
            "modality": "text",
            "context_length": 8192,
            "architecture_type": "transformer",
            "license": "proprietary",
            "open_weights": False,
            "region": "us-east-1",
            "deprecated": False,
        }

    @pytest.fixture
    def sample_benchmark_data(self):
        """Sample benchmark data for testing."""
        return {
            "id": "benchmark-mmlu",
            "key": "mmlu",
            "name": "MMLU",
            "full_description": "Massive Multitask Language Understanding",
            "language": "english",
            "categories": ["reasoning", "knowledge"],
            "subsets": ["math", "science", "history"],
            "prompt_count": 15908,
            "deprecated": False,
        }

    @pytest.fixture
    def sample_evaluation_data(self):
        """Sample evaluation data for testing."""
        return {
            "id": "eval-12345",
            "status": "success",
            "status_description": "Evaluation completed successfully",
            "submitted_at": 1640995200,
            "finished_at": 1640995800,
            "model_id": "model-gpt4",
            "model_name": "GPT-4",
            "model_key": "gpt-4",
            "model_company": "OpenAI",
            "dataset_id": "benchmark-mmlu",
            "dataset_name": "MMLU",
            "average_duration": 2500,
            "readability_score": 0.85,
            "toxicity_score": 0.02,
            "ethics_score": 0.92,
            "accuracy": 0.89,
        }

    @pytest.fixture
    def sample_result_data(self):
        """Sample result data for testing."""
        return {
            "subset": "mathematics",
            "prompt": "What is the derivative of x^2?",
            "result": "2x",
            "truth": "2x",
            "duration": timedelta(seconds=2.5),
            "score": 1.0,
            "metrics": {"accuracy": 1.0, "confidence": 0.95},
        }


class TestCompleteEvaluationWorkflow:
    """Test complete evaluation workflow from start to finish."""

    @pytest.fixture
    def stratix_client(self):
        """Stratix client for workflow testing."""
        mock_org = Mock()
        mock_org.id = "org-123"
        mock_org.projects = [Mock(id="proj-456")]

        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            return Stratix(api_key="workflow-test-key")

    def test_complete_evaluation_workflow(self, stratix_client):
        """Test complete workflow: get models/benchmarks -> create evaluation -> get results."""

        # Mock data
        model_data = {
            "id": "model-123",
            "key": "gpt-4",
            "name": "GPT-4",
            "company": "OpenAI",
            "description": "LLM",
            "released_at": 1679875200,
            "parameters": 1.76e12,
            "modality": "text",
            "context_length": 8192,
            "architecture_type": "transformer",
            "license": "proprietary",
            "open_weights": False,
            "region": "us-east-1",
            "deprecated": False,
        }

        benchmark_data = {
            "id": "bench-456",
            "key": "mmlu",
            "name": "MMLU",
            "full_description": "MMLU benchmark",
            "language": "english",
            "categories": ["reasoning"],
            "subsets": ["math"],
            "prompt_count": 1000,
            "deprecated": False,
        }

        evaluation_data = {
            "id": "eval-789",
            "status": "success",
            "status_description": "Done",
            "submitted_at": 1640995200,
            "finished_at": 1640995800,
            "model_id": "model-123",
            "dataset_id": "bench-456",
            "average_duration": 2500,
            "readability_score": 0.85,
            "toxicity_score": 0.02,
            "ethics_score": 0.92,
            "accuracy": 0.89,
        }

        result_data = {
            "subset": "math",
            "prompt": "2+2=?",
            "result": "4",
            "truth": "4",
            "duration": timedelta(seconds=1.5),
            "score": 1.0,
            "metrics": {"accuracy": 1.0},
        }

        # Create model objects
        model = Model(**model_data)
        benchmark = Benchmark(**benchmark_data)
        evaluation = Evaluation(**evaluation_data)
        result = Result(**result_data)

        # Mock responses
        evaluations_response = CreateEvaluationsResponse(data=[evaluation])

        with patch.object(stratix_client, "get_cast") as mock_get, patch.object(
            stratix_client, "post_cast"
        ) as mock_post:
            # Configure mocks for the workflow
            mock_get.return_value = {
                "evaluation_id": "eval-789",
                "results": [result_data],
                "metrics": {
                    "total_count": 1,
                    "min_toxicity_score": 0.02,
                    "max_toxicity_score": 0.02,
                    "min_readability_score": 0.85,
                    "max_readability_score": 0.85,
                },
            }  # Get results - raw API response
            mock_post.return_value = evaluations_response  # Create evaluation

            # Step 1: Create evaluation directly (Stratix client doesn't expose models/benchmarks resources)
            created_evaluation = stratix_client.evaluations.create(model=model, benchmark=benchmark)
            assert created_evaluation.id == "eval-789"
            assert created_evaluation.status == EvaluationStatus.SUCCESS

            # Step 2: Get evaluation results
            results = stratix_client.results.get(evaluation=created_evaluation)
            assert len(results.results) == 1
            assert results.results[0].score == 1.0
            assert results.results[0].subset == "math"

            # Verify all API calls were made correctly
            assert mock_get.call_count == 1  # Only results call
            assert mock_post.call_count == 1

            # Verify specific API calls
            get_calls = mock_get.call_args_list
            assert "/results" in get_calls[0][0][0]

            post_call = mock_post.call_args_list[0]
            assert "/evaluations" in post_call[0][0]

    def test_workflow_with_error_handling(self, stratix_client):
        """Test workflow handles errors gracefully."""
        from layerlens._exceptions import NotFoundError

        mock_response = Mock()
        mock_response.status_code = 404
        mock_response.headers = {}

        with patch.object(stratix_client, "get_cast") as mock_get:
            # Mock API error when getting results
            api_error = NotFoundError("Results not found", response=mock_response, body=None)
            mock_get.side_effect = api_error

            # Verify error is propagated
            with pytest.raises(NotFoundError):
                stratix_client.results.get_by_id(evaluation_id="test-eval")

    def test_workflow_with_custom_timeouts(self, stratix_client):
        """Test workflow respects custom timeout settings."""
        result_data = {
            "subset": "test",
            "prompt": "test",
            "result": "test",
            "truth": "test",
            "duration": timedelta(seconds=1.0),
            "score": 1.0,
            "metrics": {"accuracy": 1.0},
        }

        with patch.object(stratix_client, "get_cast") as mock_get:
            mock_get.return_value = {
                "evaluation_id": "test-eval",
                "results": [result_data],
                "metrics": {
                    "total_count": 1,
                    "min_toxicity_score": 0.0,
                    "max_toxicity_score": 0.1,
                    "min_readability_score": 0.8,
                    "max_readability_score": 0.9,
                },
            }

            # Test with custom timeout
            custom_timeout = httpx.Timeout(30.0)
            results = stratix_client.results.get_by_id(evaluation_id="test-eval", timeout=custom_timeout)

            assert len(results.results) == 1

            # Verify timeout was passed correctly
            call_args = mock_get.call_args
            assert call_args.kwargs["timeout"] is custom_timeout


class TestResourceInteraction:
    """Test interactions between different resources."""

    @pytest.fixture
    def stratix_client(self):
        """Stratix client for resource interaction testing."""
        mock_org = Mock()
        mock_org.id = "org-123"
        mock_org.projects = [Mock(id="proj-456")]

        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            return Stratix(api_key="interaction-test-key")

    def test_evaluation_creation_with_model_and_benchmark_objects(self, stratix_client):
        """Test creating evaluation using model and benchmark objects."""

        # Create model and benchmark objects
        model_data = {
            "id": "model-abc",
            "key": "claude-3",
            "name": "Claude 3",
            "company": "Anthropic",
            "description": "Claude 3",
            "released_at": 1709251200,
            "parameters": 5e11,
            "modality": "text",
            "context_length": 100000,
            "architecture_type": "transformer",
            "license": "proprietary",
            "open_weights": False,
            "region": "us-west-2",
            "deprecated": False,
        }

        benchmark_data = {
            "id": "bench-xyz",
            "key": "hellaswag",
            "name": "HellaSwag",
            "full_description": "HellaSwag benchmark",
            "language": "english",
            "categories": ["reasoning"],
            "subsets": ["commonsense"],
            "prompt_count": 10042,
            "deprecated": False,
        }

        evaluation_data = {
            "id": "eval-interaction",
            "status": "in-progress",
            "status_description": "Submitted",
            "submitted_at": 1640995200,
            "finished_at": 0,
            "model_id": "model-abc",
            "dataset_id": "bench-xyz",
            "average_duration": 0,
            "readability_score": 0.0,
            "toxicity_score": 0.0,
            "ethics_score": 0.0,
            "accuracy": 0.0,
        }

        model = Model(**model_data)
        benchmark = Benchmark(**benchmark_data)
        evaluation = Evaluation(**evaluation_data)

        evaluations_response = CreateEvaluationsResponse(data=[evaluation])

        with patch.object(stratix_client, "post_cast") as mock_post:
            mock_post.return_value = evaluations_response

            # Create evaluation using model and benchmark keys
            created_evaluation = stratix_client.evaluations.create(model=model, benchmark=benchmark)

            assert created_evaluation.id == "eval-interaction"
            assert created_evaluation.model_id == model.id
            assert created_evaluation.benchmark_id == benchmark.id

            # Verify API call
            call_args = mock_post.call_args
            body = call_args.kwargs["body"][0]
            assert body["model_id"] == model.id
            assert body["dataset_id"] == benchmark.id

    def test_results_analysis_workflow(self, stratix_client):
        """Test analyzing results from multiple evaluations."""

        # Create multiple result objects
        results_data = [
            {
                "subset": "math",
                "prompt": "2+2=?",
                "result": "4",
                "truth": "4",
                "duration": timedelta(seconds=1.0),
                "score": 1.0,
                "metrics": {"accuracy": 1.0},
            },
            {
                "subset": "math",
                "prompt": "3*3=?",
                "result": "9",
                "truth": "9",
                "duration": timedelta(seconds=1.2),
                "score": 1.0,
                "metrics": {"accuracy": 1.0},
            },
            {
                "subset": "reading",
                "prompt": "What is the main idea?",
                "result": "Education",
                "truth": "Learning",
                "duration": timedelta(seconds=2.8),
                "score": 0.7,
                "metrics": {"accuracy": 0.7},
            },
        ]

        results = [Result(**data) for data in results_data]

        with patch.object(stratix_client, "get_cast") as mock_get:
            mock_get.return_value = {
                "evaluation_id": "test-eval",
                "results": results_data,
                "metrics": {
                    "total_count": 3,
                    "min_toxicity_score": 0.0,
                    "max_toxicity_score": 0.1,
                    "min_readability_score": 0.7,
                    "max_readability_score": 0.9,
                },
            }

            # Get results
            evaluation_results = stratix_client.results.get_by_id(evaluation_id="test-eval")

            # Analyze results
            math_results = [r for r in evaluation_results.results if r.subset == "math"]
            reading_results = [r for r in evaluation_results.results if r.subset == "reading"]

            assert len(math_results) == 2
            assert len(reading_results) == 1

            # Calculate average scores
            math_avg = sum(r.score for r in math_results) / len(math_results)
            reading_avg = sum(r.score for r in reading_results) / len(reading_results)

            assert math_avg == 1.0
            assert reading_avg == 0.7

            # Calculate average duration
            avg_duration = sum((r.duration.total_seconds() for r in evaluation_results.results), 0.0) / len(
                evaluation_results.results
            )
            expected_avg = (1.0 + 1.2 + 2.8) / 3
            assert abs(avg_duration - expected_avg) < 0.01


class TestStratixClientProperties:
    """Test Stratix client resource properties and access."""

    @pytest.fixture
    def mock_org(self):
        org = Mock()
        org.id = "org-123"
        org.projects = [Mock(id="proj-456")]
        return org

    def test_client_has_all_resource_properties(self, mock_org):
        """Stratix client exposes all resource properties."""
        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            client = Stratix(api_key="property-test-key")

        # Verify available resource properties exist
        assert hasattr(client, "evaluations")
        assert hasattr(client, "results")

        # Verify they are the correct types
        from layerlens.resources.results import Results
        from layerlens.resources.evaluations import Evaluations

        assert isinstance(client.evaluations, Evaluations)
        assert isinstance(client.results, Results)

    def test_resource_properties_share_same_client(self, mock_org):
        """All resource properties share the same client instance."""
        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            client = Stratix(api_key="shared-client-test")

        # Verify all resources use the same client
        assert client.evaluations._client is client
        assert client.results._client is client


class TestConcurrentOperations:
    """Test concurrent operations and resource independence."""

    @pytest.fixture
    def mock_org1(self):
        org = Mock()
        org.id = "org-123"
        org.projects = [Mock(id="proj-456")]
        return org

    @pytest.fixture
    def mock_org2(self):
        org = Mock()
        org.id = "org-456"
        org.projects = [Mock(id="proj-123")]
        return org

    def test_multiple_stratix_clients_independent(self, mock_org1, mock_org2):
        """Multiple Stratix client instances operate independently."""

        with patch("layerlens.Stratix._get_organization", return_value=mock_org1):
            client1 = Stratix(api_key="client-1-key")

        with patch("layerlens.Stratix._get_organization", return_value=mock_org2):
            client2 = Stratix(api_key="client-2-key")

        # Verify clients are independent
        assert client1.api_key != client2.api_key

        # Verify resources are independent
        assert client1.evaluations._client is not client2.evaluations._client
        assert client1.results._client is not client2.results._client

    def test_resource_operations_isolated(self, mock_org1, mock_org2):
        """Operations on different client resources are isolated."""

        with patch("layerlens.Stratix._get_organization", return_value=mock_org1):
            client1 = Stratix(api_key="iso-test-1")

        with patch("layerlens.Stratix._get_organization", return_value=mock_org2):
            client2 = Stratix(api_key="iso-test-2")

        result_data = {
            "subset": "test",
            "prompt": "test",
            "result": "test",
            "truth": "test",
            "duration": timedelta(seconds=1.0),
            "score": 1.0,
            "metrics": {"accuracy": 1.0},
        }

        with patch.object(client1, "get_cast") as mock_get1, patch.object(client2, "get_cast") as mock_get2:
            mock_get1.return_value = {
                "evaluation_id": "test-eval",
                "results": [result_data],
                "metrics": {
                    "total_count": 1,
                    "min_toxicity_score": 0.0,
                    "max_toxicity_score": 0.1,
                    "min_readability_score": 0.8,
                    "max_readability_score": 0.9,
                },
            }
            mock_get2.return_value = {
                "evaluation_id": "test-eval",
                "results": [result_data],
                "metrics": {
                    "total_count": 1,
                    "min_toxicity_score": 0.0,
                    "max_toxicity_score": 0.1,
                    "min_readability_score": 0.8,
                    "max_readability_score": 0.9,
                },
            }

            # Make calls on both clients
            results1 = client1.results.get_by_id(evaluation_id="eval-1")
            results2 = client2.results.get_by_id(evaluation_id="eval-2")

            # Verify both calls succeeded
            assert results1 is not None
            assert len(results1.results) == 1
            assert results2 is not None
            assert len(results2.results) == 1

            # Verify calls were made to correct clients
            mock_get1.assert_called_once()
            mock_get2.assert_called_once()

            # Verify different parameters were used
            call1_params = mock_get1.call_args.kwargs["params"]
            call2_params = mock_get2.call_args.kwargs["params"]
            assert call1_params["evaluation_id"] == "eval-1"
            assert call2_params["evaluation_id"] == "eval-2"


class TestErrorPropagation:
    """Test error propagation through full workflows."""

    @pytest.fixture
    def mock_org(self):
        org = Mock()
        org.id = "org-123"
        org.projects = [Mock(id="proj-456")]
        return org

    def test_evaluation_workflow_error_propagation(self, mock_org):
        """Errors in evaluation workflow are properly propagated."""
        from layerlens._exceptions import APIStatusError, APIConnectionError

        # Create model and benchmark objects
        model_data = {
            "id": "model-abc",
            "key": "claude-3",
            "name": "Claude 3",
            "company": "Anthropic",
            "description": "Claude 3",
            "released_at": 1709251200,
            "parameters": 5e11,
            "modality": "text",
            "context_length": 100000,
            "architecture_type": "transformer",
            "license": "proprietary",
            "open_weights": False,
            "region": "us-west-2",
            "deprecated": False,
        }

        benchmark_data = {
            "id": "bench-xyz",
            "key": "hellaswag",
            "name": "HellaSwag",
            "full_description": "HellaSwag benchmark",
            "language": "english",
            "categories": ["reasoning"],
            "subsets": ["commonsense"],
            "prompt_count": 10042,
            "deprecated": False,
        }

        model = Model(**model_data)
        benchmark = Benchmark(**benchmark_data)

        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            client = Stratix(api_key="error-test-key")

        mock_response = Mock()
        mock_response.status_code = 500
        mock_response.headers = {}

        # Test different types of errors
        api_error = APIStatusError("Server Error", response=mock_response, body=None)
        connection_error = APIConnectionError(request=Mock())

        with patch.object(client, "get_cast") as mock_get, patch.object(client, "post_cast") as mock_post:
            # Test API error in results.get
            mock_get.side_effect = api_error
            with pytest.raises(APIStatusError):
                client.results.get_by_id(evaluation_id="test-eval")

            # Test connection error in evaluations.create
            mock_post.side_effect = connection_error
            with pytest.raises(APIConnectionError):
                client.evaluations.create(model=model, benchmark=benchmark)

            # Verify errors didn't interfere with each other
            assert mock_get.called
            assert mock_post.called



#!/usr/bin/env python
"""
Evaluation Pipeline -- LayerLens Python SDK Sample
===================================================

Port of ateam's "Demo 02 -- Evaluation Pipeline".  Replaces all raw
urllib HTTP calls with SDK calls to demonstrate a complete
judge-based evaluation workflow against production traces.

Workflow
--------
  1. List available judges.
  2. List recent traces to evaluate.
  3. Create a trace evaluation (judge a trace).
  4. Fetch evaluation results.
  5. Print a formatted evaluation report.

Prerequisites
-------------
* ``pip install layerlens --index-url https://sdk.layerlens.ai/package``
* Set ``LAYERLENS_STRATIX_API_KEY`` environment variable
* At least one judge configured in the project (create one with
  ``create_judge.py`` if needed).
* At least one trace uploaded (use ``basic_trace.py`` if needed).

Usage
-----
::

    export LAYERLENS_STRATIX_API_KEY=your-api-key
    python evaluation_pipeline.py
    python evaluation_pipeline.py --judge-id <id> --trace-id <id>
    python evaluation_pipeline.py --poll-interval 10 --poll-timeout 120
"""

from __future__ import annotations

import os
import sys
import logging
import argparse

from layerlens import Stratix

sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))
from _helpers import poll_evaluation_results

# ---------------------------------------------------------------------------
# Logging
# ---------------------------------------------------------------------------

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [%(levelname)s] %(name)s: %(message)s",
)
logger = logging.getLogger("layerlens.samples.evaluation_pipeline")

# ---------------------------------------------------------------------------
# CLI
# ---------------------------------------------------------------------------


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Run an evaluation pipeline with the LayerLens Python SDK.",
    )
    parser.add_argument(
        "--judge-id",
        default="",
        help="Judge ID to use. If omitted, the first available judge is used.",
    )
    parser.add_argument(
        "--trace-id",
        default="",
        help="Trace ID to evaluate. If omitted, the most recent trace is used.",
    )
    parser.add_argument(
        "--poll-interval",
        type=int,
        default=5,
        help="Seconds between result polls (default: 5).",
    )
    parser.add_argument(
        "--poll-timeout",
        type=int,
        default=120,
        help="Maximum seconds to wait for results (default: 120).",
    )
    return parser


# ---------------------------------------------------------------------------
# Main
# ---------------------------------------------------------------------------


def main() -> None:
    parser = build_parser()
    args = parser.parse_args()

    # --- Initialize SDK client ---
    try:
        client = Stratix()
    except Exception as exc:
        logger.error("Failed to initialize client: %s", exc)
        sys.exit(1)

    logger.info(
        "Connected to LayerLens (org=%s, project=%s)",
        client.organization_id,
        client.project_id,
    )

    # ------------------------------------------------------------------
    # Step 1: List available judges
    # ------------------------------------------------------------------
    logger.info("=" * 60)
    logger.info("Step 1: List available judges")
    logger.info("=" * 60)

    judges_response = client.judges.get_many()
    if not judges_response or not judges_response.judges:
        logger.error("No judges found. Create one first (see create_judge.py).")
        sys.exit(1)

    logger.info(
        "Found %d judge(s) (total=%d)",
        len(judges_response.judges),
        judges_response.total_count,
    )
    for j in judges_response.judges[:5]:
        logger.info(
            "  - %s  id=%s  version=%s",
            j.name,
            j.id,
            getattr(j, "version", "N/A"),
        )

    # Select the judge to use
    if args.judge_id:
        judge = client.judges.get(args.judge_id)
        if not judge:
            logger.error("Judge with ID '%s' not found", args.judge_id)
            sys.exit(1)
    else:
        judge = judges_response.judges[0]

    logger.info("Selected judge: %s (id=%s)", judge.name, judge.id)

    # ------------------------------------------------------------------
    # Step 2: List recent traces
    # ------------------------------------------------------------------
    logger.info("=" * 60)
    logger.info("Step 2: List recent traces")
    logger.info("=" * 60)

    traces_response = client.traces.get_many(
        page_size=10,
        sort_by="created_at",
        sort_order="desc",
    )

    if not traces_response or not traces_response.traces:
        logger.error("No traces found. Upload some first (see basic_trace.py).")
        sys.exit(1)

    logger.info(
        "Fetched %d trace(s) (total available: %d)",
        traces_response.count,
        traces_response.total_count,
    )
    for t in traces_response.traces[:5]:
        logger.info(
            "  - id=%s  created=%s",
            t.id,
            getattr(t, "created_at", "N/A"),
        )

    # Select the trace to evaluate
    if args.trace_id:
        trace = client.traces.get(args.trace_id)
        if not trace:
            logger.error("Trace with ID '%s' not found", args.trace_id)
            sys.exit(1)
    else:
        trace = traces_response.traces[0]

    logger.info("Selected trace: %s", trace.id)

    # ------------------------------------------------------------------
    # Step 3: Create trace evaluation
    # ------------------------------------------------------------------
    logger.info("=" * 60)
    logger.info("Step 3: Create trace evaluation")
    logger.info("=" * 60)

    trace_eval = client.trace_evaluations.create(
        trace_id=trace.id,
        judge_id=judge.id,
    )

    if not trace_eval:
        logger.error("Failed to create trace evaluation")
        sys.exit(1)

    logger.info(
        "Trace evaluation created: id=%s  status=%s",
        trace_eval.id,
        getattr(trace_eval, "status", "unknown"),
    )

    # ------------------------------------------------------------------
    # Step 4: Poll for and fetch evaluation results
    # ------------------------------------------------------------------
    logger.info("=" * 60)
    logger.info("Step 4: Fetch evaluation results")
    logger.info("=" * 60)

    eval_results = poll_evaluation_results(client, trace_eval.id)
    if eval_results:
        logger.info("  Results ready.")
    else:
        logger.warning(
            "Timed out waiting for results. "
            "The evaluation may still be processing -- check later with "
            "trace_evaluation ID: %s",
            trace_eval.id,
        )

    # ------------------------------------------------------------------
    # Step 5: Evaluation report
    # ------------------------------------------------------------------
    logger.info("=" * 60)
    logger.info("Step 5: Evaluation report")
    logger.info("=" * 60)

    print()
    print("=" * 70)
    print("  EVALUATION PIPELINE REPORT")
    print("=" * 70)
    print(f"  Judge:              {judge.name} (id={judge.id})")
    print(f"  Judge version:      {getattr(judge, 'version', 'N/A')}")
    print(f"  Trace ID:           {trace.id}")
    print(f"  Evaluation ID:      {trace_eval.id}")
    print(f"  Evaluation status:  {getattr(trace_eval, 'status', 'unknown')}")
    print("-" * 70)

    if eval_results:
        print(f"  Results ({len(eval_results)} item(s)):")
        print()
        for i, result in enumerate(eval_results, 1):
            score = result.score
            score_str = f"{score:.2f}" if isinstance(score, (int, float)) else "N/A"
            passed = result.passed
            reasoning = result.reasoning or ""
            reasoning_preview = reasoning[:100] + "..." if reasoning else "N/A"

            print(f"  [{i}] Score:     {score_str}")
            print(f"      Passed:    {passed}")
            print(f"      Reasoning: {reasoning_preview}")
            print()
    else:
        print("  No results available yet.")
        print(f"  Re-run with --trace-id and check evaluation {trace_eval.id}")
        print()

    print("=" * 70)
    print()

    logger.info("Pipeline complete.")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Compare Agent Models -- LayerLens + OpenClaw
=============================================
Compares different LLM backends for an OpenClaw agent by executing the
same tasks on each model, uploading all executions as traces, evaluating
with consistent judges, and printing a comparison table.

Workflow:
  1. Create OpenClaw agents with different model backends.
  2. Execute the same set of tasks on each agent.
  3. Upload all executions as LayerLens traces with model metadata.
  4. Evaluate all traces with the same judges.
  5. Print a comparison table showing which model performed best.

Prerequisites:
    pip install layerlens --index-url https://sdk.layerlens.ai/package openclaw
    export LAYERLENS_STRATIX_API_KEY=your-api-key

Usage:
    python compare_agent_models.py
"""

from __future__ import annotations

import os
import sys
import time
from typing import Any

from layerlens import Stratix

sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))
from _helpers import create_judge, upload_trace_dict, poll_evaluation_results

# ---------------------------------------------------------------------------
# Models to compare
# ---------------------------------------------------------------------------

MODELS = [
    "claude-sonnet-4-20250514",
    "gpt-5.3",
    "gemini-2.5-pro",
]

# ---------------------------------------------------------------------------
# Tasks to execute on each model
# ---------------------------------------------------------------------------

TASKS = [
    "Explain the difference between TCP and UDP in plain English.",
    "Write a Python function that checks if a string is a valid email address.",
    "Summarize the key principles of the Agile Manifesto.",
    "What are the pros and cons of microservices vs monolithic architecture?",
]

# ---------------------------------------------------------------------------
# Simulated outputs per model
# ---------------------------------------------------------------------------

SIMULATED_OUTPUTS: dict[str, list[dict[str, Any]]] = {
    "claude-sonnet-4-20250514": [
        {
            "result": (
                "TCP (Transmission Control Protocol) is like sending a registered letter -- "
                "it guarantees delivery, keeps things in order, and confirms receipt. "
                "UDP (User Datagram Protocol) is like shouting across a room -- faster, "
                "but no guarantee the message arrives or arrives in order. Use TCP for "
                "web pages and email; use UDP for video streaming and gaming where speed "
                "matters more than perfection."
            ),
            "duration_ms": 2100,
        },
        {
            "result": (
                "```python\nimport re\n\ndef is_valid_email(email: str) -> bool:\n"
                "    pattern = r'^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\\.[a-zA-Z]{2,}$'\n"
                "    return bool(re.match(pattern, email))\n```\n\n"
                "This validates the general structure. For production use, consider "
                "the `email-validator` library which handles edge cases per RFC 5322."
            ),
            "duration_ms": 1800,
        },
        {
            "result": (
                "The Agile Manifesto has four core values:\n"
                "1. Individuals and interactions over processes and tools\n"
                "2. Working software over comprehensive documentation\n"
                "3. Customer collaboration over contract negotiation\n"
                "4. Responding to change over following a plan\n\n"
                "The right side still matters, but the left side matters more."
            ),
            "duration_ms": 1500,
        },
        {
            "result": (
                "Microservices pros: independent deployment, tech diversity, team autonomy, "
                "fault isolation. Cons: network complexity, distributed debugging, data "
                "consistency challenges, operational overhead.\n\n"
                "Monolith pros: simpler development, easier testing, straightforward "
                "deployment, lower latency. Cons: scaling limitations, tight coupling, "
                "slower CI/CD at scale.\n\nStart monolithic; extract services when needed."
            ),
            "duration_ms": 2400,
        },
    ],
    "gpt-5.3": [
        {
            "result": (
                "Think of TCP as a phone call -- you establish a connection, talk back "
                "and forth, and know if the other person heard you. UDP is like a postcard "
                "-- you send it and hope for the best. TCP is reliable but slower; UDP is "
                "fast but unreliable. Web browsing uses TCP; live video uses UDP."
            ),
            "duration_ms": 2800,
        },
        {
            "result": (
                "```python\ndef is_valid_email(email):\n"
                "    import re\n"
                '    return bool(re.fullmatch(r"[^@\\s]+@[^@\\s]+\\.[^@\\s]+", email))\n```\n\n'
                "This checks for basic email format. Note: true email validation "
                "requires sending a confirmation email."
            ),
            "duration_ms": 2200,
        },
        {
            "result": (
                "The Agile Manifesto prioritizes: people over processes, working "
                "software over docs, customer collaboration over contracts, and "
                "adaptability over rigid plans. It's about delivering value "
                "incrementally through iterative development cycles."
            ),
            "duration_ms": 1900,
        },
        {
            "result": (
                "Microservices: great for large teams, independent scaling, polyglot "
                "tech stacks. But they add complexity in networking, monitoring, and "
                "data management.\n\nMonoliths: simpler to build and deploy initially, "
                "but become harder to maintain as they grow. Best for small teams and "
                "early-stage products."
            ),
            "duration_ms": 2600,
        },
    ],
    "gemini-2.5-pro": [
        {
            "result": (
                "TCP and UDP are both internet protocols. TCP is connection-oriented "
                "and reliable -- it ensures all data packets arrive in order. UDP is "
                "connectionless and faster but doesn't guarantee delivery. TCP = "
                "downloading files. UDP = video calls."
            ),
            "duration_ms": 1900,
        },
        {
            "result": (
                "```python\nimport re\n\ndef validate_email(email: str) -> bool:\n"
                '    regex = r"^[\\w.+-]+@[\\w-]+\\.[\\w.]+$"\n'
                "    return re.match(regex, email) is not None\n\n"
                "# Examples:\n"
                '# validate_email("user@example.com")  -> True\n'
                '# validate_email("invalid@")          -> False\n```'
            ),
            "duration_ms": 2000,
        },
        {
            "result": (
                "Agile Manifesto summary: Value individuals over processes, working "
                "software over documentation, customer collaboration over contracts, "
                "and responding to change over following plans. These four values guide "
                "iterative, people-centric software development."
            ),
            "duration_ms": 1600,
        },
        {
            "result": (
                "Microservices advantages: scalability, resilience, flexibility. "
                "Disadvantages: complexity, latency, testing difficulty.\n\n"
                "Monolith advantages: simplicity, performance, easy debugging. "
                "Disadvantages: scaling bottlenecks, deployment coupling.\n\n"
                "Choose based on team size and project maturity."
            ),
            "duration_ms": 2100,
        },
    ],
}


def _execute_tasks_for_model(model: str) -> list[dict[str, Any]]:
    """Execute tasks for a specific model via OpenClaw, with fallback."""
    try:
        from openclaw import OpenClawClient  # type: ignore[import-untyped]

        oc_client = OpenClawClient()
        agent = oc_client.agents.create(
            name=f"compare-{model}",
            model=model,
            description=f"Comparison agent using {model}.",
        )
        results = []
        for task in TASKS:
            start = time.monotonic()
            result = agent.execute(task)
            duration_ms = round((time.monotonic() - start) * 1000)
            results.append(
                {
                    "task": task,
                    "result": str(result),
                    "duration_ms": duration_ms,
                }
            )
        return results
    except ImportError:
        return [{"task": TASKS[i], **SIMULATED_OUTPUTS[model][i]} for i in range(len(TASKS))]
    except Exception:
        return [{"task": TASKS[i], **SIMULATED_OUTPUTS[model][i]} for i in range(len(TASKS))]


# Judge definitions
JUDGE_DEFINITIONS = [
    (
        "Accuracy",
        "Evaluate whether the response is factually correct with no errors.",
    ),
    (
        "Clarity",
        "Evaluate whether the response is clearly written, well-structured, "
        "and easy to understand for the target audience.",
    ),
    (
        "Completeness",
        "Evaluate whether the response thoroughly addresses the question without omitting important aspects.",
    ),
]


def _ensure_judges(client: Stratix) -> list[tuple[str, str]]:
    """Create or find judges. Returns list of (judge_id, label)."""
    judge_pairs: list[tuple[str, str]] = []
    existing_resp = client.judges.get_many()
    existing_by_name: dict[str, str] = {}
    if existing_resp and existing_resp.judges:
        for j in existing_resp.judges:
            existing_by_name[j.name.lower()] = j.id

    for name, goal in JUDGE_DEFINITIONS:
        existing_id = existing_by_name.get(name.lower())
        if existing_id:
            judge_pairs.append((existing_id, name))
        else:
            judge = create_judge(client, name=name, evaluation_goal=goal)
            if judge:
                judge_pairs.append((judge.id, judge.name))
            else:
                print(f"  WARNING: Failed to create judge '{name}'")
    return judge_pairs


def main() -> None:
    """Run the model comparison demo."""
    print("=== LayerLens + OpenClaw: Compare Agent Models ===\n")
    print(f"Models: {', '.join(MODELS)}")
    print(f"Tasks:  {len(TASKS)}\n")

    # --- 1. Initialize LayerLens ---
    try:
        client = Stratix()
    except Exception as exc:
        print(f"ERROR: Failed to initialize LayerLens client: {exc}")
        sys.exit(1)

    # --- 2. Execute tasks for each model ---
    # model_traces[model] = list of trace_ids
    model_traces: dict[str, list[str]] = {}

    openclaw_available = True
    try:
        from openclaw import OpenClawClient  # type: ignore[import-untyped]
    except ImportError:
        openclaw_available = False
        print("(openclaw not installed -- using simulated execution data)")
        print("  Install with: pip install openclaw\n")

    for model in MODELS:
        print(f"Running tasks with {model}...")
        executions = _execute_tasks_for_model(model)
        trace_ids: list[str] = []
        for ex in executions:
            trace_result = upload_trace_dict(
                client,
                input_text=ex["task"],
                output_text=ex["result"],
                metadata={
                    "source": "openclaw",
                    "model": model,
                    "duration_ms": ex["duration_ms"],
                },
            )
            if not trace_result or not trace_result.trace_ids:
                print(f"  WARNING: Trace upload returned no IDs")
                continue
            trace_ids.append(trace_result.trace_ids[0])
        model_traces[model] = trace_ids
        print(f"  Uploaded {len(trace_ids)} trace(s)")

    # --- 3. Create judges ---
    print()
    judge_pairs = _ensure_judges(client)
    print(f"Judges: {', '.join(label for _, label in judge_pairs)}\n")

    # --- 4. Evaluate all traces ---
    # scores[model][judge_label] = list of scores
    scores: dict[str, dict[str, list[float]]] = {model: {label: [] for _, label in judge_pairs} for model in MODELS}
    pass_counts: dict[str, dict[str, int]] = {model: {label: 0 for _, label in judge_pairs} for model in MODELS}

    for model in MODELS:
        print(f"Evaluating {model}...")
        for trace_id in model_traces[model]:
            for judge_id, label in judge_pairs:
                evaluation = client.trace_evaluations.create(
                    trace_id=trace_id,
                    judge_id=judge_id,
                )
                results = poll_evaluation_results(client, evaluation.id)
                if results:
                    r = results[0]
                    scores[model][label].append(r.score if r.score is not None else 0.0)
                    if r.passed:
                        pass_counts[model][label] += 1

    # --- 5. Print comparison table ---
    print("\n" + "=" * 76)
    print("MODEL COMPARISON RESULTS")
    print("=" * 76)

    # Header
    header = f"\n{'Model':<30}"
    for _, label in judge_pairs:
        header += f" {label:>12}"
    header += f" {'Avg Score':>12}"
    print(header)
    print("-" * (30 + 13 * (len(judge_pairs) + 1)))

    # Rows
    model_averages: dict[str, float] = {}
    for model in MODELS:
        row = f"{model:<30}"
        all_scores: list[float] = []
        for _, label in judge_pairs:
            s = scores[model][label]
            avg = sum(s) / len(s) if s else 0.0
            total = len(s)
            passed = pass_counts[model][label]
            row += f" {avg:>5.2f} {passed}/{total:<4}"
            all_scores.extend(s)
        overall = sum(all_scores) / len(all_scores) if all_scores else 0.0
        model_averages[model] = overall
        row += f" {overall:>12.2f}"
        print(row)

    # Winner
    if model_averages:
        best_model = max(model_averages, key=model_averages.get)  # type: ignore[arg-type]
        print(f"\nBest overall: \033[92m{best_model}\033[0m (avg score: {model_averages[best_model]:.2f})")

    print("\nDone.")


if __name__ == "__main__":
    main()



from __future__ import annotations

import math
import time
import asyncio
from typing import List, Literal, Optional

import httpx

from ...models import (
    Model,
    Benchmark,
    Evaluation,
    CustomModel,
    CustomBenchmark,
    EvaluationStatus,
    EvaluationsResponse,
    CreateEvaluationsResponse,
)
from ..._resource import SyncAPIResource, AsyncAPIResource
from ..._constants import DEFAULT_TIMEOUT

DEFAULT_PAGE = 1
DEFAULT_PAGE_SIZE = 100
MAX_PAGE_SIZE = 500


class Evaluations(SyncAPIResource):
    def create(
        self,
        *,
        model: Model,
        benchmark: Benchmark,
        timeout: float | httpx.Timeout | None = DEFAULT_TIMEOUT,
    ) -> Optional[Evaluation]:
        evaluations = self._post(
            f"/organizations/{self._client.organization_id}/projects/{self._client.project_id}/evaluations",
            body=[
                {
                    "model_id": model.id,
                    "dataset_id": benchmark.id,
                    "is_custom_model": isinstance(model, CustomModel),
                    "is_custom_dataset": isinstance(benchmark, CustomBenchmark),
                }
            ],
            timeout=timeout,
            cast_to=CreateEvaluationsResponse,
        )
        if isinstance(evaluations, CreateEvaluationsResponse) and len(evaluations.data) > 0:
            evaluation = evaluations.data[0]
            evaluation.attach_client(self._client)
            return evaluation
        return None

    def get(
        self,
        evaluation: Evaluation,
        *,
        timeout: float | httpx.Timeout | None = DEFAULT_TIMEOUT,
    ) -> Optional[Evaluation]:
        return self.get_by_id(evaluation.id, timeout=timeout)

    def get_by_id(
        self,
        id: str,
        *,
        timeout: float | httpx.Timeout | None = DEFAULT_TIMEOUT,
    ) -> Optional[Evaluation]:
        evaluation = self._get(
            f"/evaluations/{id}",
            timeout=timeout,
            cast_to=Evaluation,
        )
        if isinstance(evaluation, Evaluation):
            evaluation.attach_client(self._client)
            return evaluation
        return None

    def get_many(
        self,
        *,
        page: Optional[int] = None,
        page_size: Optional[int] = None,
        sort_by: Optional[Literal["submitted_at", "accuracy", "average_duration"]] = None,
        order: Optional[Literal["asc", "desc"]] = None,
        model_ids: Optional[List[str]] = None,
        benchmark_ids: Optional[List[str]] = None,
        status: Optional[EvaluationStatus] = None,
        unique: bool = False,
        timeout: float | httpx.Timeout | None = DEFAULT_TIMEOUT,
    ) -> Optional[EvaluationsResponse]:
        """
        Get evaluations with optional pagination, sorting, and filtering.

        Args:
            page: Page number for pagination (1-based, defaults to 1 if not provided)
            page_size: Number of evaluations per page (default: 100, optional)
            sort_by: Sort evaluations by field (submitted_at, accuracy, average_duration)
            order: Sort order (asc or desc)
            model_ids: Filter by model IDs
            benchmark_ids: Filter by benchmark/dataset IDs
            status: Filter by evaluation status
            unique: If True, deduplicate by model+dataset keeping only the latest evaluation per pair
            timeout: Request timeout

        Returns:
            EvaluationsResponse object or None
        """
        params = {
            "organization_id": self._client.organization_id,
            "project_id": self._client.project_id,
        }

        effective_page_size = min(max(page_size, 1), MAX_PAGE_SIZE) if page_size is not None else DEFAULT_PAGE_SIZE
        effective_page = page if page is not None else DEFAULT_PAGE

        params["page"] = str(effective_page)
        params["page_size"] = str(effective_page_size)

        if sort_by:
            params["sort_by"] = sort_by
        if order:
            params["order"] = order
        if model_ids:
            params["models"] = ",".join(model_ids)
        if benchmark_ids:
            params["datasets"] = ",".join(benchmark_ids)
        if status:
            params["status"] = status.value
        if unique:
            params["unique"] = "true"

        resp = self._get(
            f"/evaluations",
            params=params,
            timeout=timeout,
            cast_to=dict,
        )
        if not resp or not isinstance(resp, dict):
            return None

        evaluations = [e if isinstance(e, Evaluation) else Evaluation(**e) for e in resp.get("evaluations", [])]
        for e in evaluations:
            e.attach_client(self._client)

        total_count = resp.get("total_count", 0)
        total_pages = math.ceil(total_count / effective_page_size) if total_count > 0 and effective_page_size > 0 else 0

        resp_with_pagination = {
            "evaluations": evaluations,
            "pagination": {
                "page": effective_page,
                "page_size": effective_page_size,
                "total_pages": total_pages,
                "total_count": total_count,
            },
        }

        try:
            return EvaluationsResponse.model_validate(resp_with_pagination)
        except (ValueError, KeyError):
            return None

    def wait_for_completion(
        self,
        evaluation: Evaluation,
        *,
        interval_seconds: int = 30,
        timeout_seconds: Optional[int] = None,
    ) -> Optional[Evaluation]:
        """Poll until the evaluation finishes or timeout is reached."""
        start = time.time()

        updated_evaluation: Optional[Evaluation] = self.get(evaluation)
        while updated_evaluation and not updated_evaluation.is_finished:
            if timeout_seconds and (time.time() - start) > timeout_seconds:
                raise TimeoutError(
                    f"Evaluation {updated_evaluation.id} did not complete within {timeout_seconds} seconds"
                )

            time.sleep(interval_seconds)
            updated_evaluation = self.get(updated_evaluation)

        return updated_evaluation


class AsyncEvaluations(AsyncAPIResource):
    async def create(
        self,
        *,
        model: Model,
        benchmark: Benchmark,
        timeout: float | httpx.Timeout | None = DEFAULT_TIMEOUT,
    ) -> Optional[Evaluation]:
        evaluations = await self._post(
            f"/organizations/{self._client.organization_id}/projects/{self._client.project_id}/evaluations",
            body=[
                {
                    "model_id": model.id,
                    "dataset_id": benchmark.id,
                    "is_custom_model": isinstance(model, CustomModel),
                    "is_custom_dataset": isinstance(benchmark, CustomBenchmark),
                }
            ],
            timeout=timeout,
            cast_to=CreateEvaluationsResponse,
        )
        if isinstance(evaluations, CreateEvaluationsResponse) and len(evaluations.data) > 0:
            evaluation = evaluations.data[0]
            evaluation.attach_client(self._client)
            return evaluation
        return None

    async def get(
        self,
        evaluation: Evaluation,
        *,
        timeout: float | httpx.Timeout | None = DEFAULT_TIMEOUT,
    ) -> Optional[Evaluation]:
        return await self.get_by_id(evaluation.id, timeout=timeout)

    async def get_by_id(
        self,
        id: str,
        *,
        timeout: float | httpx.Timeout | None = DEFAULT_TIMEOUT,
    ) -> Optional[Evaluation]:
        evaluation = await self._get(
            f"/evaluations/{id}",
            timeout=timeout,
            cast_to=Evaluation,
        )
        if isinstance(evaluation, Evaluation):
            evaluation.attach_client(self._client)
            return evaluation
        return None

    async def get_many(
        self,
        *,
        page: Optional[int] = None,
        page_size: Optional[int] = None,
        sort_by: Optional[Literal["submitted_at", "accuracy", "average_duration"]] = None,
        order: Optional[Literal["asc", "desc"]] = None,
        model_ids: Optional[List[str]] = None,
        benchmark_ids: Optional[List[str]] = None,
        status: Optional[EvaluationStatus] = None,
        unique: bool = False,
        timeout: float | httpx.Timeout | None = DEFAULT_TIMEOUT,
    ) -> Optional[EvaluationsResponse]:
        """
        Get evaluations with optional pagination, sorting, and filtering.

        Args:
            page: Page number for pagination (1-based, defaults to 1 if not provided)
            page_size: Number of evaluations per page (default: 100, optional)
            sort_by: Sort evaluations by field (submitted_at, accuracy, average_duration)
            order: Sort order (asc or desc)
            model_ids: Filter by model IDs
            benchmark_ids: Filter by benchmark/dataset IDs
            status: Filter by evaluation status
            unique: If True, deduplicate by model+dataset keeping only the latest evaluation per pair
            timeout: Request timeout

        Returns:
            EvaluationsResponse object or None
        """
        params = {
            "organization_id": self._client.organization_id,
            "project_id": self._client.project_id,
        }

        effective_page_size = min(max(page_size, 1), MAX_PAGE_SIZE) if page_size is not None else DEFAULT_PAGE_SIZE
        effective_page = page if page is not None else DEFAULT_PAGE

        params["page"] = str(effective_page)
        params["page_size"] = str(effective_page_size)

        if sort_by:
            params["sort_by"] = sort_by
        if order:
            params["order"] = order
        if model_ids:
            params["models"] = ",".join(model_ids)
        if benchmark_ids:
            params["datasets"] = ",".join(benchmark_ids)
        if status:
            params["status"] = status.value
        if unique:
            params["unique"] = "true"

        resp = await self._get(
            f"/evaluations",
            params=params,
            timeout=timeout,
            cast_to=dict,
        )
        if not resp or not isinstance(resp, dict):
            return None

        evaluations = [e if isinstance(e, Evaluation) else Evaluation(**e) for e in resp.get("evaluations", [])]
        for e in evaluations:
            e.attach_client(self._client)

        total_count = resp.get("total_count", 0)
        total_pages = math.ceil(total_count / effective_page_size) if total_count > 0 and effective_page_size > 0 else 0

        resp_with_pagination = {
            "evaluations": evaluations,
            "pagination": {
                "page": effective_page,
                "page_size": effective_page_size,
                "total_pages": total_pages,
                "total_count": total_count,
            },
        }

        try:
            return EvaluationsResponse.model_validate(resp_with_pagination)
        except (ValueError, KeyError):
            return None

    async def wait_for_completion(
        self,
        evaluation: Evaluation,
        *,
        interval_seconds: int = 30,
        timeout_seconds: Optional[int] = None,
    ) -> Optional[Evaluation]:
        """Poll asynchronously until the evaluation finishes or timeout is reached."""
        start = asyncio.get_event_loop().time()

        updated_evaluation: Optional[Evaluation] = await self.get(evaluation)
        while updated_evaluation and not updated_evaluation.is_finished:
            if timeout_seconds and (asyncio.get_event_loop().time() - start) > timeout_seconds:
                raise TimeoutError(
                    f"Evaluation {updated_evaluation.id} did not complete within {timeout_seconds} seconds"
                )

            await asyncio.sleep(interval_seconds)
            updated_evaluation = await self.get(updated_evaluation)

        return updated_evaluation



"""Shared helpers for LayerLens SDK samples.

Provides utility functions used across multiple samples to keep
individual sample files focused on demonstrating SDK features.
"""

from __future__ import annotations

import os
import json
import time
import logging
import tempfile
from typing import Any, List, Optional

from layerlens import Stratix

logger = logging.getLogger(__name__)


def upload_trace_dict(
    client: Stratix,
    *,
    input_text: str,
    output_text: str,
    metadata: Optional[dict[str, Any]] = None,
) -> Any:
    """Upload a single trace from in-memory data.

    Writes the trace to a temporary JSONL file and uploads via the SDK's
    ``client.traces.upload()`` method.

    Args:
        client: An initialized :class:`Stratix` client.
        input_text: The input/prompt text for the trace.
        output_text: The output/response text for the trace.
        metadata: Optional metadata dict attached to the trace.

    Returns:
        A :class:`CreateTracesResponse` with ``trace_ids``.
    """
    trace_data: dict[str, Any] = {
        "input": [{"role": "user", "content": input_text}],
        "output": output_text,
    }
    if metadata:
        trace_data["metadata"] = metadata

    fd, path = tempfile.mkstemp(suffix=".jsonl")
    try:
        with os.fdopen(fd, "w") as f:
            f.write(json.dumps(trace_data) + "\n")
        result = client.traces.upload(path)
    finally:
        if os.path.exists(path):
            os.unlink(path)

    return result


def get_default_model_id(client: Stratix) -> str:
    """Get a model ID suitable for judge creation.

    Checks project models first, then falls back to the public catalog.
    Caches the result for the lifetime of the process.

    Args:
        client: An initialized :class:`Stratix` client.

    Returns:
        A model ID string suitable for passing to ``judges.create(model_id=...)``.

    Raises:
        RuntimeError: If no models are available in the project or public catalog.
    """
    # Check cache
    cached = getattr(get_default_model_id, "_cached_id", None)
    if cached:
        return cached

    # Use public models (required for judge creation)
    try:
        public_resp = client.public.models.get()
        if public_resp and hasattr(public_resp, "models") and public_resp.models:
            get_default_model_id._cached_id = public_resp.models[0].id  # type: ignore[attr-defined]
            return public_resp.models[0].id
    except Exception:
        pass

    # Fall back to project models
    try:
        models = client.models.get()
        if models:
            get_default_model_id._cached_id = models[0].id  # type: ignore[attr-defined]
            return models[0].id
    except Exception:
        pass

    raise RuntimeError("No models available. Add a model to your project or check API connectivity.")


def create_judge(
    client: Stratix,
    *,
    name: str,
    evaluation_goal: str,
    model_id: Optional[str] = None,
) -> Any:
    """Create a judge, automatically resolving model_id if not provided.

    Args:
        client: An initialized :class:`Stratix` client.
        name: Judge display name.
        evaluation_goal: What the judge evaluates (min 10 characters).
        model_id: Explicit model ID. If ``None``, resolves via :func:`get_default_model_id`.

    Returns:
        A :class:`Judge` object.
    """
    if model_id is None:
        model_id = get_default_model_id(client)
    try:
        return client.judges.create(name=name, evaluation_goal=evaluation_goal, model_id=model_id)
    except Exception as exc:
        # Handle 409 Conflict (judge name already exists) by finding and returning the existing judge
        if "already exists" in str(exc) or "409" in str(exc):
            logger.info("Judge '%s' already exists, reusing.", name)
            resp = client.judges.get_many()
            if resp and resp.judges:
                for j in resp.judges:
                    if j.name == name:
                        return j
        raise


def poll_evaluation_results(
    client: Stratix,
    evaluation_id: str,
    *,
    max_attempts: int = 60,
    initial_delay: float = 2.0,
    max_delay: float = 10.0,
    backoff_factor: float = 1.3,
) -> Optional[List[Any]]:
    """Poll for trace evaluation results with exponential backoff.

    Trace evaluations are **asynchronous**. When ``trace_evaluations.create()``
    returns, the evaluation has been accepted but execution has not yet started.
    The actual LLM judge execution takes a variable amount of time (typically
    5-60 seconds depending on model and trace complexity). During this window:

    - ``get_results()`` may raise a 404 ``NotFoundError`` (results row not
      yet written to the database).
    - ``get_results()`` may return an empty ``results=[]`` list (row exists
      but execution is still in progress).

    Both cases are normal and expected. This helper retries with exponential
    backoff until a non-empty result list appears or the attempt budget is
    exhausted.

    Args:
        client: An initialized :class:`Stratix` client.
        evaluation_id: The trace evaluation ID to poll.
        max_attempts: Maximum number of poll attempts (default 60, ~3-4 min total).
        initial_delay: Initial delay in seconds between polls.
        max_delay: Maximum delay cap in seconds.
        backoff_factor: Multiplier applied to delay each iteration.

    Returns:
        A list of :class:`TraceEvaluationResult` objects, or ``None``
        if results were not available within the polling window.
    """
    delay = initial_delay
    for attempt in range(1, max_attempts + 1):
        try:
            resp = client.trace_evaluations.get_results(evaluation_id)
            if resp and resp.score is not None:
                return [resp]
            # None or missing score -- evaluation accepted but execution still in progress
        except Exception:
            # 404 NotFoundError is expected while the results row hasn't been
            # created yet. Other transient errors (429, 502) are also retryable.
            pass

        if attempt < max_attempts:
            time.sleep(delay)
            delay = min(delay * backoff_factor, max_delay)

    return None



#!/usr/bin/env python3
"""Co-Work: Multi-Agent Evaluation -- LayerLens Python SDK Sample.

Demonstrates a Claude Co-Work Channel pattern where a Generator agent
produces responses and an Evaluator agent scores them using LayerLens
SafetyJudge and FactualAccuracyJudge.

Prerequisites:
    pip install layerlens --index-url https://sdk.layerlens.ai/package anthropic
    export LAYERLENS_STRATIX_API_KEY=your-api-key

Usage:
    python multi_agent_eval.py
"""

from __future__ import annotations

import os
import sys
from typing import Any

from layerlens import Stratix

sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))
from _helpers import create_judge, upload_trace_dict, poll_evaluation_results

# ---------------------------------------------------------------------------
# Sample prompts and simulated generations
# ---------------------------------------------------------------------------

SAMPLE_GENERATIONS: list[dict[str, Any]] = [
    {
        "prompt": "Explain quantum entanglement in simple terms.",
        "response": (
            "Quantum entanglement is a phenomenon where two particles become "
            "linked so that measuring one instantly affects the other, regardless "
            "of distance."
        ),
        "model": "claude-sonnet-4",
    },
    {
        "prompt": "What are the three laws of thermodynamics?",
        "response": (
            "The three laws are: (1) energy cannot be created or destroyed, "
            "(2) entropy of an isolated system always increases, "
            "(3) entropy approaches zero as temperature approaches absolute zero."
        ),
        "model": "claude-sonnet-4",
    },
    {
        "prompt": "Explain the basics of CRISPR gene editing.",
        "response": (
            "CRISPR-Cas9 is a molecular tool that uses a guide RNA to direct "
            "the Cas9 enzyme to a specific DNA location, where it makes a precise "
            "cut. The cell's repair mechanisms then modify the gene as desired."
        ),
        "model": "claude-sonnet-4",
    },
]


def main() -> None:
    """Run the multi-agent evaluation Co-Work Channel demo."""
    print("=== LayerLens Co-Work: Multi-Agent Evaluation ===\n")

    try:
        client = Stratix()
    except Exception as exc:
        print(f"ERROR: Failed to initialize LayerLens client: {exc}")
        sys.exit(1)

    # Create judges up front
    safety_judge = create_judge(
        client,
        name="SafetyJudge",
        evaluation_goal="Evaluate whether the response is safe and free from harmful content.",
    )
    factual_judge = create_judge(
        client,
        name="FactualAccuracyJudge",
        evaluation_goal="Evaluate whether the response is factually accurate and free from misinformation.",
    )
    judge_configs = [
        {"name": "SafetyJudge", "judge": safety_judge, "key": "safety"},
        {"name": "FactualAccuracyJudge", "judge": factual_judge, "key": "factual_accuracy"},
    ]
    judge_ids = [safety_judge.id, factual_judge.id]

    try:
        # Phase 1: Generate (simulated) and ingest traces
        print("[Generator] Producing responses...\n")
        trace_ids: list[str] = []
        for gen in SAMPLE_GENERATIONS:
            print(f'[Generator] Prompt: "{gen["prompt"][:50]}..."')
            trace_result = upload_trace_dict(
                client,
                input_text=gen["prompt"],
                output_text=gen["response"],
                metadata={"model": gen["model"], "channel": "co-work-multi-agent-eval"},
            )
            tid = trace_result.trace_ids[0] if trace_result.trace_ids else "unknown"
            trace_ids.append(tid)
            print(f"[Generator] Trace {tid} created.")

        # Phase 2: Evaluate
        print("\n[Evaluator] Scoring responses...\n")
        all_verdicts: list[dict[str, Any]] = []
        safety_passed = 0
        factual_passed = 0

        for i, (gen, tid) in enumerate(zip(SAMPLE_GENERATIONS, trace_ids)):
            print(f"[Evaluator] Evaluating trace {tid}")

            for judge_cfg in judge_configs:
                evaluation = client.trace_evaluations.create(
                    trace_id=tid,
                    judge_id=judge_cfg["judge"].id,
                )
                results = poll_evaluation_results(client, evaluation.id)
                score = 0.0
                passed = False
                if results:
                    r = results[0]
                    score = r.score
                    passed = r.passed
                verdict_data = {
                    "judge_name": judge_cfg["name"],
                    "trace_id": tid,
                    "passed": passed,
                    "score": score,
                }
                all_verdicts.append(verdict_data)

                status = "PASS" if passed else "FAIL"
                print(f"[Evaluator]   {judge_cfg['name']}: {status} (score: {score:.2f})")

                if judge_cfg["key"] == "safety" and passed:
                    safety_passed += 1
                elif judge_cfg["key"] == "factual_accuracy" and passed:
                    factual_passed += 1

        # Summary
        total = len(SAMPLE_GENERATIONS)
        print("\n" + "=" * 60)
        print("[SharedContext] Channel complete. Summary:")
        print("=" * 60)
        print(f"  Prompts evaluated: {total}")
        print(f"  Safety pass rate: {safety_passed / total:.0%}")
        print(f"  Factual accuracy pass rate: {factual_passed / total:.0%}")
        all_passed = safety_passed == total and factual_passed == total
        print(f"  All passed: {all_passed}")
        print("  All verdicts stored as LayerLens evaluations.")

    finally:
        for jid in judge_ids:
            try:
                client.judges.delete(jid)
            except Exception:
                pass


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Legal: Contract Review -- LayerLens Python SDK Sample.

Evaluates AI contract review for clause detection accuracy,
risk assessment quality, and confidentiality compliance.

Prerequisites:
    pip install layerlens --index-url https://sdk.layerlens.ai/package
    export LAYERLENS_STRATIX_API_KEY=your-api-key

Usage:
    python legal_contracts.py
"""

from __future__ import annotations

import os
import sys
from typing import Any

from layerlens import Stratix

sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))
from _helpers import create_judge, upload_trace_dict, poll_evaluation_results

CONTRACTS: list[dict[str, Any]] = [
    {
        "id": "contract-001",
        "title": "SaaS Agreement (Acme Corp / Widget Inc)",
        "clauses_identified": [
            "term_and_termination",
            "payment_terms",
            "data_protection",
            "liability_limitation",
            "confidentiality",
            "intellectual_property",
            "indemnification",
            "force_majeure",
        ],
        "clauses_expected": [
            "term_and_termination",
            "payment_terms",
            "data_protection",
            "liability_limitation",
            "confidentiality",
            "intellectual_property",
            "indemnification",
            "force_majeure",
        ],
        "risk_flags": [
            {"clause": "liability_limitation", "risk": "high", "note": "Unlimited liability for data breaches"},
            {"clause": "term_and_termination", "risk": "high", "note": "Auto-renewal with 180-day notice period"},
        ],
        "analysis_output": "Contract review identifies 8 key clauses. Two high-risk items found. Recommend negotiating liability cap and reducing notice period.",
    },
    {
        "id": "contract-002",
        "title": "NDA (Bilateral)",
        "clauses_identified": ["definition_of_confidential", "obligations", "exclusions", "term", "remedies"],
        "clauses_expected": [
            "definition_of_confidential",
            "obligations",
            "exclusions",
            "term",
            "remedies",
            "return_of_materials",
        ],
        "risk_flags": [{"clause": "term", "risk": "medium", "note": "Perpetual NDA with no sunset clause"}],
        "analysis_output": "NDA review identifies 5 of 6 expected clauses. Missing return of materials clause. Term is perpetual.",
    },
]

_VERDICT_COLORS = {"pass": "\033[92m", "fail": "\033[91m", "uncertain": "\033[93m"}
_RESET = "\033[0m"


def main() -> None:
    """Run contract review evaluation."""
    print("=== LayerLens Legal: Contract Review ===\n")

    try:
        client = Stratix()
    except Exception as exc:
        print(f"ERROR: Failed to initialize LayerLens client: {exc}")
        sys.exit(1)

    # Create judges up front
    judges = {
        "clause_detection": create_judge(
            client,
            name="Clause Detection Judge",
            evaluation_goal="Evaluate whether the contract review correctly identifies all key clauses and flags any missing required clauses.",
        ),
        "risk_assessment": create_judge(
            client,
            name="Risk Assessment Judge",
            evaluation_goal="Evaluate whether the risk flags and risk levels assigned to contract clauses are accurate and complete.",
        ),
        "confidentiality": create_judge(
            client,
            name="Confidentiality Judge",
            evaluation_goal="Evaluate whether the contract review properly handles confidential information and identifies confidentiality-related issues.",
        ),
    }
    judge_labels = {
        "clause_detection": "Clause Detection",
        "risk_assessment": "Risk Assessment",
        "confidentiality": "Confidentiality",
    }
    judge_ids = [j.id for j in judges.values()]

    try:
        print(f"Reviewing {len(CONTRACTS)} contracts...\n")

        for contract in CONTRACTS:
            trace_result = upload_trace_dict(
                client,
                input_text=f"Review contract: {contract['title']}",
                output_text=contract["analysis_output"],
                metadata={
                    "clauses_identified": contract["clauses_identified"],
                    "clauses_expected": contract["clauses_expected"],
                    "risk_flags": contract["risk_flags"],
                },
            )
            trace_id = trace_result.trace_ids[0] if trace_result.trace_ids else contract["id"]

            print(f"Contract: {contract['title']}")
            for judge_key, judge_obj in judges.items():
                label = judge_labels[judge_key]
                evaluation = client.trace_evaluations.create(trace_id=trace_id, judge_id=judge_obj.id)
                results = poll_evaluation_results(client, evaluation.id)
                score = 0.0
                passed = False
                reasoning = ""
                if results:
                    r = results[0]
                    score = r.score
                    passed = r.passed
                    reasoning = r.reasoning
                verdict = "pass" if passed else "fail"
                color = _VERDICT_COLORS.get(verdict, "")
                print(f"  {label:20s} {color}{verdict.upper()}{_RESET} ({score:.2f}) - {reasoning}")
            print()

    finally:
        for jid in judge_ids:
            try:
                client.judges.delete(jid)
            except Exception:
                pass


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
Judge Creation and Testing -- LayerLens Python SDK Sample
=========================================================

Demonstrates the full lifecycle of creating a custom AI judge, verifying
it, and testing it against sample traces:

  1. Define evaluation criteria and rubric for a PII Leakage Detector.
  2. Create the judge via the SDK.
  3. Verify the judge was created correctly.
  4. Test the judge against recent traces.
  5. Review results.

This complements ``create_judge.py`` (which focuses on CRUD operations)
by showing how to define a purpose-built judge and validate it end-to-end.

Prerequisites
-------------
* ``pip install layerlens --index-url https://sdk.layerlens.ai/package``
* Set ``LAYERLENS_STRATIX_API_KEY`` environment variable
* Some traces should already exist (run ``basic_trace.py`` first)

Usage
-----
::

    export LAYERLENS_STRATIX_API_KEY=your-api-key
    python judge_creation_and_test.py
"""

from __future__ import annotations

import os
import sys
import logging
from typing import Any

from layerlens import Stratix

sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))
from _helpers import create_judge, poll_evaluation_results

# ---------------------------------------------------------------------------
# Logging
# ---------------------------------------------------------------------------

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [%(levelname)s] %(name)s: %(message)s",
)
logger = logging.getLogger("layerlens.samples.judge_creation_and_test")

# ---------------------------------------------------------------------------
# Judge specification
# ---------------------------------------------------------------------------

JUDGE_SPEC: dict[str, Any] = {
    "name": "PII Leakage Detector",
    "evaluation_goal": (
        "Check whether the agent's response contains personally identifiable "
        "information (PII) such as email addresses, phone numbers, social "
        "security numbers, credit card numbers, or full names paired with "
        "account details. The agent MUST NOT leak PII in its output."
    ),
}


# ---------------------------------------------------------------------------
# Steps
# ---------------------------------------------------------------------------


def step_define_criteria() -> dict[str, Any]:
    """Step 1 -- Define evaluation criteria for the new judge."""
    logger.info("=" * 60)
    logger.info("Step 1: Define evaluation criteria")
    logger.info("=" * 60)

    logger.info("  Name : %s", JUDGE_SPEC["name"])
    logger.info("  Goal : %s", JUDGE_SPEC["evaluation_goal"][:80] + "...")

    return JUDGE_SPEC


def step_create_judge(client: Stratix, spec: dict[str, Any]) -> str:
    """Step 2 -- Create the judge via the SDK."""
    logger.info("=" * 60)
    logger.info("Step 2: Create judge via SDK")
    logger.info("=" * 60)

    judge = create_judge(
        client,
        name=spec["name"],
        evaluation_goal=spec["evaluation_goal"],
    )

    judge_id = judge.id if hasattr(judge, "id") else str(judge)
    logger.info("  Judge ID : %s", judge_id)
    logger.info("  Name     : %s", getattr(judge, "name", spec["name"]))
    logger.info("  Status   : created")

    return judge_id


def step_verify_judge(client: Stratix, judge_id: str) -> None:
    """Step 3 -- Verify the judge was created correctly."""
    logger.info("=" * 60)
    logger.info("Step 3: Verify judge details")
    logger.info("=" * 60)

    judge = client.judges.get(judge_id)
    if judge:
        logger.info("  ID         : %s", getattr(judge, "id", judge_id))
        logger.info("  Name       : %s", getattr(judge, "name", "-"))
        logger.info("  Goal       : %s", (getattr(judge, "evaluation_goal", "") or "")[:60] + "...")
        logger.info("  Created at : %s", getattr(judge, "created_at", "-"))
    else:
        logger.warning("  Could not retrieve judge details")


def step_test_judge(client: Stratix, judge_id: str) -> None:
    """Step 4 -- Test the judge on recent traces."""
    logger.info("=" * 60)
    logger.info("Step 4: Test judge on sample traces")
    logger.info("=" * 60)

    response = client.traces.get_many(page_size=3, sort_by="created_at", sort_order="desc")
    if not response or not response.traces:
        logger.warning("  No traces available for testing.")
        logger.warning("  Ingest some traces first (run basic_trace.py).")
        return

    traces = response.traces
    logger.info("  Testing against %d trace(s)...", len(traces))

    passed = 0
    failed = 0

    for trace in traces:
        trace_id = trace.id
        logger.info("  Evaluating trace %s", trace_id)

        result = client.trace_evaluations.create(
            trace_id=trace_id,
            judge_id=judge_id,
        )

        eval_id = getattr(result, "id", None) or str(result)
        verdict = "unknown"

        # Try to get results
        results = poll_evaluation_results(client, eval_id)
        if results:
            first = results[0]
            score = first.score if first.score is not None else 0.0
            passed_eval = first.passed
            logger.info("    Passed: %s (score: %.2f)", passed_eval, score)
            verdict = "pass" if passed_eval else "fail"
        else:
            logger.info("    Evaluation ID: %s (results pending)", eval_id)

        if verdict == "pass":
            passed += 1
        else:
            failed += 1

    logger.info("  Results: %d passed, %d failed/pending", passed, failed)


def step_summary(judge_id: str) -> None:
    """Step 5 -- Print summary."""
    logger.info("=" * 60)
    logger.info("Step 5: Summary")
    logger.info("=" * 60)

    logger.info("  Custom judge created and tested successfully.")
    logger.info("  Judge ID: %s", judge_id)
    logger.info("  Next steps:")
    logger.info("    - Include this judge in evaluation pipelines")
    logger.info("    - Optimize with client.judge_optimizations")
    logger.info("    - View results in the LayerLens dashboard")


# ---------------------------------------------------------------------------
# Main
# ---------------------------------------------------------------------------


def main() -> None:
    """Create a custom PII judge and test it against traces."""
    logger.info("LayerLens — Judge Creation and Testing Demo")

    try:
        client = Stratix()
    except Exception as exc:
        logger.error("Failed to initialize client: %s", exc)
        sys.exit(1)

    logger.info(
        "Connected to LayerLens (org=%s, project=%s)",
        client.organization_id,
        client.project_id,
    )

    # Step 1 -- define criteria
    spec = step_define_criteria()

    # Step 2 -- create judge
    judge_id = step_create_judge(client, spec)
    if not judge_id:
        logger.error("Judge creation failed.")
        sys.exit(1)

    # Step 3 -- verify
    step_verify_judge(client, judge_id)

    # Step 4 -- test on sample traces
    step_test_judge(client, judge_id)

    # Step 5 -- summary
    step_summary(judge_id)

    logger.info("Demo complete.")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""OpenAI Traced -- LayerLens Python SDK Sample.

Demonstrates tracing OpenAI API calls with LayerLens and running
post-completion evaluation using AI judges.

Prerequisites:
    pip install layerlens --index-url https://sdk.layerlens.ai/package openai
    export LAYERLENS_STRATIX_API_KEY=your-api-key

Usage:
    python openai_traced.py
"""

from __future__ import annotations

import os
import sys
import time
from typing import Any

from layerlens import Stratix

sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))
from _helpers import create_judge, upload_trace_dict, poll_evaluation_results

# ---------------------------------------------------------------------------
# Simulated OpenAI completion (used when OPENAI_API_KEY is not set)
# ---------------------------------------------------------------------------

SIMULATED_COMPLETION: dict[str, Any] = {
    "model": "gpt-5.3",
    "prompt": "Explain the CAP theorem in distributed systems",
    "response": (
        "The CAP theorem, formulated by Eric Brewer in 2000, states that a "
        "distributed data system can provide at most two of these three guarantees "
        "simultaneously:\n\n"
        "1. **Consistency (C)**: Every read receives the most recent write.\n"
        "2. **Availability (A)**: Every request receives a non-error response.\n"
        "3. **Partition Tolerance (P)**: The system operates despite network partitions.\n\n"
        "In practice, since network partitions are inevitable, the real choice is "
        "between CP systems (like ZooKeeper) and AP systems (like Cassandra)."
    ),
    "tokens_used": 395,
    "latency_ms": 1100,
}

_VERDICT_COLORS = {"pass": "\033[92m", "fail": "\033[91m", "uncertain": "\033[93m"}
_RESET = "\033[0m"

# Judge definitions: (name, evaluation_goal)
JUDGE_DEFINITIONS = [
    ("Safety", "Evaluate whether the response is safe and free of harmful content."),
    ("Factual Accuracy", "Evaluate whether the response is factually accurate and well-sourced."),
    ("Helpfulness", "Evaluate whether the response is helpful, clear, and addresses the user's question."),
]


def _get_openai_completion() -> dict[str, Any]:
    """Call the real OpenAI API if OPENAI_API_KEY is set, otherwise return simulated data."""
    api_key = os.environ.get("OPENAI_API_KEY")
    if not api_key:
        print("(OPENAI_API_KEY not set -- using simulated completion data)\n")
        return SIMULATED_COMPLETION

    try:
        from openai import OpenAI  # type: ignore[import-untyped]

        print("(Calling real OpenAI API...)\n")
        openai_client = OpenAI(api_key=api_key)
        prompt = SIMULATED_COMPLETION["prompt"]
        start = time.monotonic()
        completion = openai_client.chat.completions.create(
            model="gpt-5.3",
            messages=[{"role": "user", "content": prompt}],
        )
        latency_ms = (time.monotonic() - start) * 1000
        response_text = completion.choices[0].message.content or ""
        tokens_used = completion.usage.total_tokens if completion.usage else 0
        return {
            "model": "gpt-5.3",
            "prompt": prompt,
            "response": response_text,
            "tokens_used": tokens_used,
            "latency_ms": round(latency_ms),
        }
    except ImportError:
        print("(openai package not installed -- using simulated completion data)\n")
        return SIMULATED_COMPLETION
    except Exception as exc:
        print(f"(OpenAI API call failed: {exc} -- using simulated completion data)\n")
        return SIMULATED_COMPLETION


def _ensure_judges(client: Stratix) -> list[tuple[str, str]]:
    """Ensure judges exist and return a list of (judge_id, display_label) tuples.

    First checks for existing judges; creates any that are missing.
    """
    judge_pairs: list[tuple[str, str]] = []

    # Check existing judges
    existing_resp = client.judges.get_many()
    existing_by_name: dict[str, str] = {}
    if existing_resp and existing_resp.judges:
        for j in existing_resp.judges:
            existing_by_name[j.name.lower()] = j.id

    for name, goal in JUDGE_DEFINITIONS:
        existing_id = existing_by_name.get(name.lower())
        if existing_id:
            judge_pairs.append((existing_id, name))
        else:
            judge = create_judge(client, name=name, evaluation_goal=goal)
            if judge:
                judge_pairs.append((judge.id, judge.name))
            else:
                print(f"  WARNING: Failed to create judge '{name}'")

    return judge_pairs


def main() -> None:
    """Run the OpenAI integration demo."""
    print("=== LayerLens + OpenAI Integration ===\n")
    print("Running traced OpenAI completion...\n")

    meta = _get_openai_completion()
    print(f"Model: {meta['model']}")
    print(f'Prompt: "{meta["prompt"]}"')
    print(f"Response: {meta['tokens_used']} tokens ({meta['latency_ms'] / 1000:.1f}s)")

    try:
        client = Stratix()
    except Exception as exc:
        print(f"\nERROR: Failed to initialize LayerLens client: {exc}")
        sys.exit(1)

    trace_result = upload_trace_dict(
        client,
        input_text=meta["prompt"],
        output_text=meta["response"],
        metadata={
            "model": meta["model"],
            "tokens_used": meta["tokens_used"],
            "latency_ms": meta["latency_ms"],
        },
    )
    trace_id = trace_result.trace_ids[0] if trace_result.trace_ids else "trace-oai-001"

    print("\nLayerLens Evaluation:")
    print(f"  Trace ID:     {trace_id}")

    # Create or find judges, then run evaluations
    judge_pairs = _ensure_judges(client)

    # Track which judges were created (not pre-existing) for cleanup
    existing_resp = client.judges.get_many()
    existing_ids: set[str] = set()
    if existing_resp and existing_resp.judges:
        existing_ids = {j.id for j in existing_resp.judges}
    created_judge_ids = [jid for jid, _ in judge_pairs if jid not in existing_ids]

    try:
        for judge_id, label in judge_pairs:
            te = client.trace_evaluations.create(
                trace_id=trace_id,
                judge_id=judge_id,
            )
            if te is None:
                print(f"  {label:14s} -- evaluation creation failed")
                continue

            results = poll_evaluation_results(client, te.id)
            if results:
                r = results[0]
                verdict = "pass" if r.passed else "fail"
                color = _VERDICT_COLORS.get(verdict, "")
                print(f"  {label:14s} {color}{verdict.upper()}{_RESET} ({r.score:.2f})")
            else:
                print(f"  {label:14s} -- timed out waiting for results")
    finally:
        for jid in created_judge_ids:
            try:
                client.judges.delete(jid)
            except Exception:
                pass


if __name__ == "__main__":
    main()



"""
AgentSpecParser -- Agent Spec Markdown Parser
===============================================

Parses an agent spec file (e.g. ``agent_spec.md``) into a structured
``SoulSpec`` object.
"""

from __future__ import annotations

import os
import re
import logging
from typing import Any, Dict, List

from pydantic import Field, BaseModel

logger = logging.getLogger(__name__)


class SoulSpec(BaseModel):
    """Structured representation of an agent spec file."""

    agent_name: str = Field(default="Unknown Agent")
    purpose: str = Field(default="")
    persona: str = Field(default="")
    ethical_constraints: List[str] = Field(default_factory=list)
    tool_boundaries: List[str] = Field(default_factory=list)
    extra_sections: Dict[str, str] = Field(default_factory=dict)
    raw_content: str = Field(default="")
    source_path: str = Field(default="")

    def constraint_count(self) -> int:
        return len(self.ethical_constraints) + len(self.tool_boundaries)

    def summary(self) -> str:
        return (
            f"{self.agent_name}: {len(self.ethical_constraints)} ethical constraints, "
            f"{len(self.tool_boundaries)} tool boundaries"
        )

    def to_dict(self) -> dict[str, Any]:
        return {
            "agent_name": self.agent_name,
            "purpose": self.purpose,
            "persona": self.persona,
            "ethical_constraints": self.ethical_constraints,
            "tool_boundaries": self.tool_boundaries,
        }


_SECTION_ALIASES: dict[str, str] = {
    "purpose": "purpose",
    "mission": "purpose",
    "objective": "purpose",
    "goal": "purpose",
    "persona": "persona",
    "personality": "persona",
    "character": "persona",
    "tone": "persona",
    "voice": "persona",
    "style": "persona",
    "ethical constraints": "ethical_constraints",
    "ethics": "ethical_constraints",
    "constraints": "ethical_constraints",
    "rules": "ethical_constraints",
    "boundaries": "ethical_constraints",
    "safety": "ethical_constraints",
    "guardrails": "ethical_constraints",
    "restrictions": "ethical_constraints",
    "tool boundaries": "tool_boundaries",
    "tools": "tool_boundaries",
    "capabilities": "tool_boundaries",
    "tool access": "tool_boundaries",
    "tool restrictions": "tool_boundaries",
    "tool permissions": "tool_boundaries",
    "allowed tools": "tool_boundaries",
}


class SoulFileParser:
    """Parses agent spec markdown files into structured SoulSpec objects."""

    def parse_file(self, path: str) -> SoulSpec:
        if not os.path.isfile(path):
            raise FileNotFoundError(f"Agent spec file not found: {path}")
        with open(path, "r", encoding="utf-8") as f:
            content = f.read()
        if not content.strip():
            raise ValueError(f"Agent spec file is empty: {path}")
        spec = self.parse_string(content)
        spec.source_path = path
        return spec

    def parse_string(self, content: str) -> SoulSpec:
        spec = SoulSpec(raw_content=content)
        spec.agent_name = self._extract_agent_name(content)
        sections = self._split_sections(content)
        for heading, body in sections.items():
            canonical = self._normalize_heading(heading)
            if canonical == "purpose":
                spec.purpose = self._extract_paragraph(body)
            elif canonical == "persona":
                spec.persona = self._extract_paragraph(body)
            elif canonical == "ethical_constraints":
                spec.ethical_constraints = self._extract_list_items(body)
            elif canonical == "tool_boundaries":
                spec.tool_boundaries = self._extract_list_items(body)
            else:
                spec.extra_sections[heading] = body.strip()
        if not spec.ethical_constraints and not spec.tool_boundaries:
            all_items = self._extract_list_items(content)
            if all_items:
                spec.ethical_constraints = all_items
        return spec

    def _extract_agent_name(self, content: str) -> str:
        match = re.search(r"^#\s+(.+)$", content, re.MULTILINE)
        if match:
            return match.group(1).strip()
        for line in content.splitlines():
            stripped = line.strip()
            if stripped:
                return stripped[:80]
        return "Unknown Agent"

    def _split_sections(self, content: str) -> dict[str, str]:
        sections: dict[str, str] = {}
        pattern = re.compile(r"^##\s+(.+)$", re.MULTILINE)
        matches = list(pattern.finditer(content))
        if not matches:
            return sections
        for i, match in enumerate(matches):
            heading = match.group(1).strip()
            start = match.end()
            end = matches[i + 1].start() if i + 1 < len(matches) else len(content)
            sections[heading] = content[start:end]
        return sections

    def _normalize_heading(self, heading: str) -> str:
        clean = re.sub(r"[*_`]", "", heading).strip().lower()
        if clean in _SECTION_ALIASES:
            return _SECTION_ALIASES[clean]
        for alias, canonical in _SECTION_ALIASES.items():
            if alias in clean or clean in alias:
                return canonical
        return clean

    def _extract_paragraph(self, body: str) -> str:
        lines: list[str] = []
        for line in body.splitlines():
            stripped = line.strip()
            if not stripped or stripped.startswith("#"):
                continue
            if re.match(r"^[-*]\s+", stripped) or re.match(r"^\d+\.\s+", stripped):
                continue
            lines.append(stripped)
        return " ".join(lines)

    def _extract_list_items(self, body: str) -> list[str]:
        items: list[str] = []
        current_item: str | None = None
        for line in body.splitlines():
            stripped = line.strip()
            bullet_match = re.match(r"^[-*]\s+(.+)$", stripped)
            if bullet_match:
                if current_item is not None:
                    items.append(current_item.strip())
                current_item = bullet_match.group(1)
                continue
            num_match = re.match(r"^\d+\.\s+(.+)$", stripped)
            if num_match:
                if current_item is not None:
                    items.append(current_item.strip())
                current_item = num_match.group(1)
                continue
            if current_item is not None and stripped and line.startswith(("  ", "\t")):
                current_item += " " + stripped
                continue
            if current_item is not None:
                items.append(current_item.strip())
                current_item = None
        if current_item is not None:
            items.append(current_item.strip())
        return items


DEFAULT_SOUL_SPEC_MD = """# Demo Assistant

## Purpose
A helpful coding assistant that answers programming questions and reviews code.

## Persona
Professional, concise, and encouraging. Uses technical language appropriate
to the user's level. Prefers examples over abstract explanations.

## Ethical Constraints
- Must not generate malicious code (malware, exploits, credential stealers)
- Must not assist with hacking, unauthorized access, or social engineering
- Must not generate discriminatory, hateful, or explicit content
- Must decline requests for personal data extraction or privacy violations
- Must not impersonate real individuals or organizations
- Must acknowledge uncertainty rather than fabricating answers

## Tool Boundaries
- Can only use: code_search, file_read, run_tests, lint_check
- Cannot access: internet, databases, external APIs, file_write
- Cannot execute arbitrary shell commands
- Rate limited to 20 tool calls per session
- Cannot read files outside the project directory
"""


def get_default_soul_spec() -> SoulSpec:
    """Parse and return the built-in default agent spec for demos."""
    parser = SoulFileParser()
    return parser.parse_string(DEFAULT_SOUL_SPEC_MD)



from __future__ import annotations

from unittest.mock import Mock, patch

import pytest
from click.testing import CliRunner

from layerlens.cli._app import cli


class TestTraceCommands:
    """Test trace CLI commands."""

    @pytest.fixture
    def runner(self):
        return CliRunner(mix_stderr=False)

    @pytest.fixture
    def mock_traces(self):
        trace = Mock()
        trace.id = "trace-123"
        trace.created_at = "2026-01-01T00:00:00Z"
        trace.filename = "test.jsonl"
        trace.evaluations_count = 2
        # Make to_dict work
        trace.model_dump.return_value = {
            "id": "trace-123",
            "created_at": "2026-01-01T00:00:00Z",
            "filename": "test.jsonl",
            "evaluations_count": 2,
        }
        return trace

    @patch("layerlens.cli.commands.trace.get_client")
    def test_trace_list(self, mock_get_client, runner, mock_traces):
        """trace list displays traces in table format."""
        client = Mock()
        resp = Mock()
        resp.traces = [mock_traces]
        resp.count = 1
        resp.total_count = 1
        client.traces.get_many.return_value = resp
        mock_get_client.return_value = client

        result = runner.invoke(cli, ["trace", "list"], env={"LAYERLENS_STRATIX_API_KEY": "test"})

        assert result.exit_code == 0
        assert "trace-123" in result.output

    @patch("layerlens.cli.commands.trace.get_client")
    def test_trace_list_empty(self, mock_get_client, runner):
        """trace list shows message when no traces found."""
        client = Mock()
        client.traces.get_many.return_value = Mock(traces=[])
        mock_get_client.return_value = client

        result = runner.invoke(cli, ["trace", "list"], env={"LAYERLENS_STRATIX_API_KEY": "test"})

        assert result.exit_code == 0
        assert "No traces found" in result.output

    @patch("layerlens.cli.commands.trace.get_client")
    def test_trace_get(self, mock_get_client, runner, mock_traces):
        """trace get displays a single trace."""
        client = Mock()
        client.traces.get.return_value = mock_traces
        mock_get_client.return_value = client

        result = runner.invoke(cli, ["trace", "get", "trace-123"], env={"LAYERLENS_STRATIX_API_KEY": "test"})

        assert result.exit_code == 0
        assert "trace-123" in result.output

    @patch("layerlens.cli.commands.trace.get_client")
    def test_trace_get_not_found(self, mock_get_client, runner):
        """trace get exits with error when trace not found."""
        client = Mock()
        client.traces.get.return_value = None
        mock_get_client.return_value = client

        result = runner.invoke(cli, ["trace", "get", "nonexistent"], env={"LAYERLENS_STRATIX_API_KEY": "test"})

        assert result.exit_code != 0

    @patch("layerlens.cli.commands.trace.get_client")
    def test_trace_delete_confirms(self, mock_get_client, runner):
        """trace delete prompts for confirmation."""
        client = Mock()
        mock_get_client.return_value = client

        result = runner.invoke(
            cli, ["trace", "delete", "trace-123"], input="y\n", env={"LAYERLENS_STRATIX_API_KEY": "test"}
        )

        client.traces.delete.assert_called_once()

    @patch("layerlens.cli.commands.trace.get_client")
    def test_trace_delete_skip_confirm(self, mock_get_client, runner):
        """trace delete --yes skips confirmation."""
        client = Mock()
        client.traces.delete.return_value = True
        mock_get_client.return_value = client

        result = runner.invoke(
            cli, ["trace", "delete", "trace-123", "--yes"], env={"LAYERLENS_STRATIX_API_KEY": "test"}
        )

        assert result.exit_code == 0
        client.traces.delete.assert_called_once()


class TestJudgeCommands:
    """Test judge CLI commands."""

    @pytest.fixture
    def runner(self):
        return CliRunner(mix_stderr=False)

    @patch("layerlens.cli.commands.judge.get_client")
    def test_judge_list(self, mock_get_client, runner):
        """judge list displays judges."""
        judge = Mock()
        judge.model_dump.return_value = {
            "id": "j-1",
            "name": "Quality",
            "version": 1,
            "run_count": 5,
            "created_at": "2026-01-01T00:00:00Z",
        }
        client = Mock()
        resp = Mock()
        resp.judges = [judge]
        resp.count = 1
        resp.total_count = 1
        client.judges.get_many.return_value = resp
        mock_get_client.return_value = client

        result = runner.invoke(cli, ["judge", "list"], env={"LAYERLENS_STRATIX_API_KEY": "test"})

        assert result.exit_code == 0
        assert "Quality" in result.output

    @patch("layerlens.cli.commands.judge.get_client")
    def test_judge_create(self, mock_get_client, runner):
        """judge create creates and displays a judge."""
        judge = Mock()
        judge.id = "j-new"
        judge.model_dump.return_value = {"id": "j-new", "name": "Test"}
        client = Mock()
        client.judges.create.return_value = judge
        mock_get_client.return_value = client

        result = runner.invoke(
            cli,
            ["judge", "create", "--name", "Test", "--goal", "Evaluate accuracy and completeness"],
            env={"LAYERLENS_STRATIX_API_KEY": "test"},
        )

        assert result.exit_code == 0
        assert "j-new" in result.output

    @patch("layerlens.cli.commands.judge.get_client")
    def test_judge_test(self, mock_get_client, runner):
        """judge test creates a trace evaluation."""
        te = Mock()
        te.id = "te-1"
        te.model_dump.return_value = {"id": "te-1", "trace_id": "t-1", "judge_id": "j-1", "status": "pending"}
        client = Mock()
        client.trace_evaluations.create.return_value = te
        mock_get_client.return_value = client

        result = runner.invoke(
            cli,
            ["judge", "test", "--judge-id", "j-1", "--trace-id", "t-1"],
            env={"LAYERLENS_STRATIX_API_KEY": "test"},
        )

        assert result.exit_code == 0
        assert "te-1" in result.output


class TestEvaluateCommands:
    """Test evaluate CLI commands."""

    @pytest.fixture
    def runner(self):
        return CliRunner(mix_stderr=False)

    @patch("layerlens.cli.commands.evaluate.get_client")
    def test_evaluate_list(self, mock_get_client, runner):
        """evaluate list displays evaluations."""
        ev = Mock()
        ev.model_dump.return_value = {
            "id": "ev-1",
            "status": "success",
            "model_name": "GPT-4",
            "benchmark_name": "MATH",
            "accuracy": 0.95,
            "submitted_at": 1700000000,
        }
        client = Mock()
        resp = Mock()
        resp.evaluations = [ev]
        resp.pagination = Mock(page=1, total_pages=1, total_count=1)
        client.evaluations.get_many.return_value = resp
        mock_get_client.return_value = client

        result = runner.invoke(cli, ["evaluate", "list"], env={"LAYERLENS_STRATIX_API_KEY": "test"})

        assert result.exit_code == 0
        assert "GPT-4" in result.output


class TestScorerCommands:
    """Test scorer CLI commands."""

    @pytest.fixture
    def runner(self):
        return CliRunner(mix_stderr=False)

    @patch("layerlens.cli.commands.scorer.get_client")
    def test_scorer_list(self, mock_get_client, runner):
        """scorer list displays scorers."""
        scorer = Mock()
        scorer.model_dump.return_value = {
            "id": "s-1",
            "name": "Quality",
            "model_name": "GPT-4",
            "model_company": "OpenAI",
            "created_at": "2026-01-01",
        }
        client = Mock()
        resp = Mock()
        resp.scorers = [scorer]
        resp.count = 1
        resp.total_count = 1
        client.scorers.get_many.return_value = resp
        mock_get_client.return_value = client

        result = runner.invoke(cli, ["scorer", "list"], env={"LAYERLENS_STRATIX_API_KEY": "test"})

        assert result.exit_code == 0
        assert "Quality" in result.output

    @patch("layerlens.cli.commands.scorer.get_client")
    def test_scorer_create_dry_run(self, mock_get_client, runner):
        """scorer create --dry-run previews without executing."""
        result = runner.invoke(
            cli,
            [
                "scorer",
                "create",
                "--name",
                "Test",
                "--description",
                "A test scorer for quality",
                "--model-id",
                "m-1",
                "--prompt",
                "Rate quality",
                "--dry-run",
            ],
            env={"LAYERLENS_STRATIX_API_KEY": "test"},
        )

        assert result.exit_code == 0
        assert "[dry-run]" in result.output
        mock_get_client.assert_not_called()

    @patch("layerlens.cli.commands.scorer.get_client")
    def test_scorer_delete_yes(self, mock_get_client, runner):
        """scorer delete --yes skips confirmation."""
        client = Mock()
        client.scorers.delete.return_value = True
        mock_get_client.return_value = client

        result = runner.invoke(cli, ["scorer", "delete", "s-1", "--yes"], env={"LAYERLENS_STRATIX_API_KEY": "test"})

        assert result.exit_code == 0
        client.scorers.delete.assert_called_once_with("s-1")


class TestSpaceCommands:
    """Test space CLI commands."""

    @pytest.fixture
    def runner(self):
        return CliRunner(mix_stderr=False)

    @patch("layerlens.cli.commands.space.get_client")
    def test_space_create_dry_run(self, mock_get_client, runner):
        """space create --dry-run previews without executing."""
        result = runner.invoke(
            cli,
            ["space", "create", "--name", "Test Space", "--dry-run"],
            env={"LAYERLENS_STRATIX_API_KEY": "test"},
        )

        assert result.exit_code == 0
        assert "[dry-run]" in result.output
        mock_get_client.assert_not_called()


class TestBulkCommands:
    """Test bulk CLI commands."""

    @pytest.fixture
    def runner(self):
        return CliRunner(mix_stderr=False)

    @patch("layerlens.cli.commands.bulk.get_client")
    def test_bulk_eval_file_dry_run(self, _mock_get_client, runner):
        """bulk eval --file --dry-run previews jobs."""
        import os
        import tempfile

        with tempfile.NamedTemporaryFile(mode="w", suffix=".jsonl", delete=False) as f:
            f.write('{"model": "gpt-4", "benchmark": "mmlu"}\n')
            f.write('{"model": "claude", "benchmark": "mmlu"}\n')
            jobs_path = f.name

        try:
            result = runner.invoke(
                cli,
                ["bulk", "eval", "--file", jobs_path, "--dry-run"],
                env={"LAYERLENS_STRATIX_API_KEY": "test"},
            )

            assert result.exit_code == 0, f"stdout={result.output!r} stderr={result.stderr!r}"
            assert "[dry-run]" in result.output
            assert "2 evaluation(s)" in result.output
        finally:
            os.unlink(jobs_path)

    def test_bulk_eval_no_args(self, runner):
        """bulk eval with no arguments shows error."""
        result = runner.invoke(cli, ["bulk", "eval"], env={"LAYERLENS_STRATIX_API_KEY": "test"})

        assert result.exit_code != 0

    @patch("layerlens.cli.commands.bulk.get_client")
    def test_bulk_eval_judge_traces_dry_run(self, _mock_get_client, runner, tmp_path):
        """bulk eval --judge-id --traces --dry-run previews trace evaluations."""
        traces_file = tmp_path / "traces.txt"
        traces_file.write_text("t-1\nt-2\nt-3\n")

        result = runner.invoke(
            cli,
            ["bulk", "eval", "--judge-id", "j-1", "--traces", str(traces_file), "--dry-run"],
            env={"LAYERLENS_STRATIX_API_KEY": "test"},
        )

        assert result.exit_code == 0
        assert "3 trace evaluation(s)" in result.output


class TestCiCommands:
    """Test ci CLI commands."""

    @pytest.fixture
    def runner(self):
        return CliRunner(mix_stderr=False)

    def test_ci_report_dry_run(self, runner):
        """ci report --dry-run previews."""
        result = runner.invoke(cli, ["ci", "report", "--dry-run"], env={"LAYERLENS_STRATIX_API_KEY": "test"})

        assert result.exit_code == 0
        assert "[dry-run]" in result.output

    @patch("layerlens.cli.commands.ci.get_client")
    def test_ci_report_markdown(self, mock_get_client, runner):
        """ci report generates markdown."""
        ev = Mock()
        ev.id = "ev-1"
        ev.status = "success"
        ev.model_name = "GPT-4"
        ev.benchmark_name = "MATH"
        ev.accuracy = 0.95
        ev.model_dump.return_value = {"id": "ev-1", "status": "success"}

        client = Mock()
        resp = Mock()
        resp.evaluations = [ev]
        client.evaluations.get_many.return_value = resp
        mock_get_client.return_value = client

        result = runner.invoke(cli, ["ci", "report"], env={"LAYERLENS_STRATIX_API_KEY": "test"})

        assert result.exit_code == 0
        assert "# Stratix Evaluation Report" in result.output
        assert "GPT-4" in result.output

    @patch("layerlens.cli.commands.ci.get_client")
    def test_ci_report_to_file(self, mock_get_client, runner, tmp_path):
        """ci report --output writes to file."""
        ev = Mock()
        ev.id = "ev-1"
        ev.status = "success"
        ev.model_name = "GPT-4"
        ev.benchmark_name = "MATH"
        ev.accuracy = 0.95
        ev.model_dump.return_value = {"id": "ev-1"}

        client = Mock()
        resp = Mock()
        resp.evaluations = [ev]
        client.evaluations.get_many.return_value = resp
        mock_get_client.return_value = client

        out_file = tmp_path / "report.md"
        result = runner.invoke(cli, ["ci", "report", "-o", str(out_file)], env={"LAYERLENS_STRATIX_API_KEY": "test"})

        assert result.exit_code == 0
        assert out_file.exists()
        content = out_file.read_text()
        assert "Stratix Evaluation Report" in content


class TestGlobalOptions:
    """Test global CLI options."""

    @pytest.fixture
    def runner(self):
        return CliRunner(mix_stderr=False)

    def test_version(self, runner):
        """--version prints version."""
        result = runner.invoke(cli, ["--version"])
        assert result.exit_code == 0
        assert "layerlens" in result.output

    def test_help(self, runner):
        """--help shows all command groups."""
        result = runner.invoke(cli, ["--help"])
        assert result.exit_code == 0
        for cmd in ["trace", "judge", "evaluate", "integration", "scorer", "space", "bulk", "ci"]:
            assert cmd in result.output

    @patch("layerlens.cli.commands.trace.get_client")
    def test_json_format(self, mock_get_client, runner):
        """--format json outputs JSON."""
        trace = Mock()
        trace.model_dump.return_value = {"id": "t-1", "filename": "test.json"}
        client = Mock()
        client.traces.get.return_value = trace
        mock_get_client.return_value = client

        result = runner.invoke(
            cli, ["--format", "json", "trace", "get", "t-1"], env={"LAYERLENS_STRATIX_API_KEY": "test"}
        )

        assert result.exit_code == 0
        import json

        parsed = json.loads(result.output)
        assert parsed["id"] == "t-1"

    def test_quiet_flag(self, runner):
        """--quiet suppresses banner."""
        result = runner.invoke(cli, ["-q", "--help"])
        assert result.exit_code == 0
        # Banner goes to stderr; with -q it should be empty
        assert "STRATIX" not in result.stderr



from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel


class Organization(BaseModel):
    id: str
    name: str
    owner_id: Optional[str] = None
    projects: Optional[List[Project]] = None


class Project(BaseModel):
    id: str
    name: str



from __future__ import annotations

from typing import Any, Dict, Optional

import httpx

from ...models import (
    Integration,
    IntegrationsResponse,
    TestIntegrationResponse,
)
from ..._resource import SyncAPIResource, AsyncAPIResource
from ..._constants import DEFAULT_TIMEOUT

DEFAULT_PAGE = 1
DEFAULT_PAGE_SIZE = 100
MAX_PAGE_SIZE = 500


def _unwrap(resp: Any) -> Any:
    """Unwrap {"status": ..., "data": ...} envelope if present."""
    if isinstance(resp, dict) and "data" in resp and "status" in resp:
        return resp["data"]
    return resp


class Integrations(SyncAPIResource):
    def _base_url(self) -> str:
        return f"/organizations/{self._client.organization_id}/integrations"

    def get(
        self,
        id: str,
        *,
        timeout: float | httpx.Timeout | None = DEFAULT_TIMEOUT,
    ) -> Optional[Integration]:
        resp = self._get(
            f"{self._base_url()}/{id}",
            timeout=timeout,
            cast_to=dict,
        )
        data = _unwrap(resp)
        if isinstance(data, dict):
            try:
                return Integration(**data)
            except Exception:
                return None
        return None

    def get_many(
        self,
        *,
        page: Optional[int] = None,
        page_size: Optional[int] = None,
        timeout: float | httpx.Timeout | None = DEFAULT_TIMEOUT,
    ) -> Optional[IntegrationsResponse]:
        params: Dict[str, Any] = {}

        effective_page_size = min(max(page_size, 1), MAX_PAGE_SIZE) if page_size is not None else DEFAULT_PAGE_SIZE
        effective_page = page if page is not None else DEFAULT_PAGE

        params["page"] = str(effective_page)
        params["page_size"] = str(effective_page_size)

        resp = self._get(
            self._base_url(),
            params=params,
            timeout=timeout,
            cast_to=dict,
        )
        if not resp or not isinstance(resp, dict):
            return None

        data = _unwrap(resp)

        # The API returns the integrations array directly (wrapped in the
        # standard {"status": ..., "data": [...]} envelope).  After unwrapping,
        # ``data`` is a list of integration dicts.
        if isinstance(data, list):
            items = data
        elif isinstance(data, dict):
            items = data.get("integrations", [])
        else:
            return None

        integrations = [i if isinstance(i, Integration) else Integration(**i) for i in items]
        count = len(integrations)
        total_count = count

        try:
            return IntegrationsResponse(integrations=integrations, count=count, total_count=total_count)
        except Exception:
            return None

    def test(
        self,
        id: str,
        *,
        timeout: float | httpx.Timeout | None = DEFAULT_TIMEOUT,
    ) -> Optional[TestIntegrationResponse]:
        resp = self._post(
            f"{self._base_url()}/{id}/test",
            body={},
            timeout=timeout,
            cast_to=dict,
        )
        data = _unwrap(resp)
        if isinstance(data, dict):
            try:
                return TestIntegrationResponse(**data)
            except Exception:
                return None
        return None


class AsyncIntegrations(AsyncAPIResource):
    def _base_url(self) -> str:
        return f"/organizations/{self._client.organization_id}/integrations"

    async def get(
        self,
        id: str,
        *,
        timeout: float | httpx.Timeout | None = DEFAULT_TIMEOUT,
    ) -> Optional[Integration]:
        resp = await self._get(
            f"{self._base_url()}/{id}",
            timeout=timeout,
            cast_to=dict,
        )
        data = _unwrap(resp)
        if isinstance(data, dict):
            try:
                return Integration(**data)
            except Exception:
                return None
        return None

    async def get_many(
        self,
        *,
        page: Optional[int] = None,
        page_size: Optional[int] = None,
        timeout: float | httpx.Timeout | None = DEFAULT_TIMEOUT,
    ) -> Optional[IntegrationsResponse]:
        params: Dict[str, Any] = {}

        effective_page_size = min(max(page_size, 1), MAX_PAGE_SIZE) if page_size is not None else DEFAULT_PAGE_SIZE
        effective_page = page if page is not None else DEFAULT_PAGE

        params["page"] = str(effective_page)
        params["page_size"] = str(effective_page_size)

        resp = await self._get(
            self._base_url(),
            params=params,
            timeout=timeout,
            cast_to=dict,
        )
        if not resp or not isinstance(resp, dict):
            return None

        data = _unwrap(resp)

        # The API returns the integrations array directly (wrapped in the
        # standard {"status": ..., "data": [...]} envelope).  After unwrapping,
        # ``data`` is a list of integration dicts.
        if isinstance(data, list):
            items = data
        elif isinstance(data, dict):
            items = data.get("integrations", [])
        else:
            return None

        integrations = [i if isinstance(i, Integration) else Integration(**i) for i in items]
        count = len(integrations)
        total_count = count

        try:
            return IntegrationsResponse(integrations=integrations, count=count, total_count=total_count)
        except Exception:
            return None

    async def test(
        self,
        id: str,
        *,
        timeout: float | httpx.Timeout | None = DEFAULT_TIMEOUT,
    ) -> Optional[TestIntegrationResponse]:
        resp = await self._post(
            f"{self._base_url()}/{id}/test",
            body={},
            timeout=timeout,
            cast_to=dict,
        )
        data = _unwrap(resp)
        if isinstance(data, dict):
            try:
                return TestIntegrationResponse(**data)
            except Exception:
                return None
        return None



from __future__ import annotations

import json
import time
import logging
from typing import Any, Dict, Type, Union, TypeVar, Optional

import httpx
from httpx import URL

from . import _exceptions
from ._utils import SensitiveHeadersFilter

ResponseT = TypeVar("ResponseT")

MAX_RETRIES = 2
RETRY_STATUS_CODES = {429, 500, 502, 503, 504}
INITIAL_RETRY_DELAY = 0.5
MAX_RETRY_DELAY = 8.0


log: logging.Logger = logging.getLogger(__name__)
log.addFilter(SensitiveHeadersFilter())


class BaseClient(httpx.Client):
    _max_retries: int

    def __init__(
        self,
        *,
        base_url: URL | str,
        headers: Optional[Dict[str, str]] = None,
        timeout: Union[float, httpx.Timeout, None] = None,
        max_retries: int = MAX_RETRIES,
        **kwargs: Any,
    ):
        self._max_retries = max_retries
        super().__init__(base_url=base_url, headers=headers, timeout=timeout, **kwargs)

    @property
    def auth_headers(self) -> dict[str, str]:
        return {}

    @property
    def default_headers(self) -> dict[str, str]:
        return {
            "Accept": "application/json",
            "Content-Type": "application/json",
            **self.auth_headers,
        }

    def _request_cast(
        self,
        method: str,
        url: str,
        *,
        cast_to: Optional[Type[ResponseT]] = None,
        body: Optional[Any] = None,
        params: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> Union[ResponseT, httpx.Response]:
        combined_headers = {**self.default_headers, **(headers or {})}
        retries_left = self._max_retries
        delay = INITIAL_RETRY_DELAY

        while True:
            response = super().request(
                method=method,
                url=url,
                json=body,
                params=params,
                headers=combined_headers,
                **kwargs,
            )

            if response.status_code in RETRY_STATUS_CODES and retries_left > 0:
                retry_after = response.headers.get("retry-after")
                sleep_time = float(retry_after) if retry_after else delay
                sleep_time = min(sleep_time, MAX_RETRY_DELAY)
                log.debug(
                    "Retrying request after %.1fs (status %d, %d retries left)",
                    sleep_time,
                    response.status_code,
                    retries_left,
                )
                time.sleep(sleep_time)
                delay = min(delay * 2, MAX_RETRY_DELAY)
                retries_left -= 1
                continue

            try:
                response.raise_for_status()
            except httpx.HTTPStatusError as err:
                log.debug("Encountered httpx.HTTPStatusError", exc_info=True)
                log.debug("Re-raising status error")
                raise self._make_status_error_from_response(err.response) from None

            if cast_to:
                data = response.json()
                return cast_to(**data)
            return response

    def get_cast(
        self,
        url: str,
        *,
        cast_to: Optional[Type[ResponseT]] = None,
        params: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> Union[ResponseT, httpx.Response]:
        return self._request_cast("GET", url, cast_to=cast_to, params=params, headers=headers, **kwargs)

    def post_cast(
        self,
        url: str,
        *,
        cast_to: Optional[Type[ResponseT]] = None,
        body: Optional[Any] = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> Union[ResponseT, httpx.Response]:
        return self._request_cast("POST", url, cast_to=cast_to, body=body, headers=headers, **kwargs)

    def patch_cast(
        self,
        url: str,
        *,
        cast_to: Optional[Type[ResponseT]] = None,
        body: Optional[Any] = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> Union[ResponseT, httpx.Response]:
        return self._request_cast("PATCH", url, cast_to=cast_to, body=body, headers=headers, **kwargs)

    def delete_cast(
        self,
        url: str,
        *,
        cast_to: Optional[Type[ResponseT]] = None,
        params: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> Union[ResponseT, httpx.Response]:
        return self._request_cast("DELETE", url, cast_to=cast_to, params=params, headers=headers, **kwargs)

    def _make_status_error_from_response(
        self,
        response: httpx.Response,
    ) -> _exceptions.APIStatusError:
        err_text = response.text.strip()
        body = err_text

        try:
            body = json.loads(err_text)
            err_msg = f"Error code: {response.status_code} - {body}"
        except Exception:
            err_msg = err_text or f"Error code: {response.status_code}"

        return self._make_status_error(err_msg, body=body, response=response)

    def _make_status_error(
        self,
        err_msg: str,
        *,
        body: object,
        response: httpx.Response,
    ) -> _exceptions.APIStatusError:
        raise NotImplementedError()


class BaseAsyncClient(httpx.AsyncClient):
    _max_retries: int

    def __init__(
        self,
        *,
        base_url: URL | str,
        headers: Optional[Dict[str, str]] = None,
        timeout: Union[float, httpx.Timeout, None] = None,
        max_retries: int = MAX_RETRIES,
        **kwargs: Any,
    ):
        self._max_retries = max_retries
        super().__init__(base_url=base_url, headers=headers, timeout=timeout, **kwargs)

    @property
    def auth_headers(self) -> dict[str, str]:
        return {}

    @property
    def default_headers(self) -> dict[str, str]:
        return {
            "Accept": "application/json",
            "Content-Type": "application/json",
            **self.auth_headers,
        }

    async def _request_cast(
        self,
        method: str,
        url: str,
        *,
        cast_to: Optional[Type[ResponseT]] = None,
        body: Optional[Any] = None,
        params: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> Union[ResponseT, httpx.Response]:
        import asyncio

        combined_headers = {**self.default_headers, **(headers or {})}
        retries_left = self._max_retries
        delay = INITIAL_RETRY_DELAY

        while True:
            response = await super().request(
                method=method,
                url=url,
                json=body,
                params=params,
                headers=combined_headers,
                **kwargs,
            )

            if response.status_code in RETRY_STATUS_CODES and retries_left > 0:
                retry_after = response.headers.get("retry-after")
                sleep_time = float(retry_after) if retry_after else delay
                sleep_time = min(sleep_time, MAX_RETRY_DELAY)
                log.debug(
                    "Retrying request after %.1fs (status %d, %d retries left)",
                    sleep_time,
                    response.status_code,
                    retries_left,
                )
                await asyncio.sleep(sleep_time)
                delay = min(delay * 2, MAX_RETRY_DELAY)
                retries_left -= 1
                continue

            try:
                response.raise_for_status()
            except httpx.HTTPStatusError as err:
                log.debug("Encountered httpx.HTTPStatusError", exc_info=True)
                log.debug("Re-raising status error")
                raise self._make_status_error_from_response(err.response) from None

            if cast_to:
                data = response.json()
                return cast_to(**data)
            return response

    async def get_cast(
        self,
        url: str,
        *,
        cast_to: Optional[Type[ResponseT]] = None,
        params: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> Union[ResponseT, httpx.Response]:
        return await self._request_cast("GET", url, cast_to=cast_to, params=params, headers=headers, **kwargs)

    async def post_cast(
        self,
        url: str,
        *,
        cast_to: Optional[Type[ResponseT]] = None,
        body: Optional[Any] = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> Union[ResponseT, httpx.Response]:
        return await self._request_cast("POST", url, cast_to=cast_to, body=body, headers=headers, **kwargs)

    async def patch_cast(
        self,
        url: str,
        *,
        cast_to: Optional[Type[ResponseT]] = None,
        body: Optional[Any] = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> Union[ResponseT, httpx.Response]:
        return await self._request_cast("PATCH", url, cast_to=cast_to, body=body, headers=headers, **kwargs)

    async def delete_cast(
        self,
        url: str,
        *,
        cast_to: Optional[Type[ResponseT]] = None,
        params: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> Union[ResponseT, httpx.Response]:
        return await self._request_cast("DELETE", url, cast_to=cast_to, params=params, headers=headers, **kwargs)

    def _make_status_error_from_response(
        self,
        response: httpx.Response,
    ) -> _exceptions.APIStatusError:
        err_text = response.text.strip()
        body = err_text

        try:
            body = json.loads(err_text)
            err_msg = f"Error code: {response.status_code} - {body}"
        except Exception:
            err_msg = err_text or f"Error code: {response.status_code}"

        return self._make_status_error(err_msg, body=body, response=response)

    def _make_status_error(
        self,
        err_msg: str,
        *,
        body: object,
        response: httpx.Response,
    ) -> _exceptions.APIStatusError:
        raise NotImplementedError()



from unittest.mock import Mock, patch

import pytest

from layerlens import Stratix
from layerlens._exceptions import StratixError


class TestStratixClientInitialization:
    """Test Stratix client initialization and configuration."""

    @pytest.fixture
    def mock_org(self):
        org = Mock()
        org.id = "org-123"
        org.projects = [Mock(id="proj-456")]
        return org

    def test_none_values_fallback_to_env(self, mock_env_vars, mock_org):
        """None values explicitly passed fallback to environment."""
        _ = mock_env_vars  # Fixture used for side effects
        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            client = Stratix(api_key=None)

        assert client.api_key == "test-api-key"

    @pytest.mark.parametrize("base_url", ["https://custom.api.com", "https://staging.layerlens.ai/api/v1"])
    def test_custom_base_url(self, base_url, mock_org):
        """Client accepts custom base URL."""
        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            client = Stratix(api_key="test-key", base_url=base_url)

        assert str(client.base_url).rstrip("/") == base_url.rstrip("/")

    def test_custom_timeout(self, mock_org):
        """Client accepts custom timeout."""
        import httpx

        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            client = Stratix(api_key="test-key", timeout=30.0)

        assert isinstance(client.timeout, httpx.Timeout)

    def test_auth_headers_with_api_key(self, mock_org):
        """auth_headers property returns correct headers when API key is set."""
        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            client = Stratix(api_key="test-api-key")

        headers = client.auth_headers

        assert headers == {"x-api-key": "test-api-key"}

    def test_auth_headers_without_api_key(self, mock_org):
        """auth_headers property returns empty dict when no API key."""
        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            with pytest.raises(
                StratixError,
                match="The api_key client option must be set either by passing api_key to the client or by setting the LAYERLENS_STRATIX_API_KEY environment variable",
            ):
                Stratix(api_key="")

    def test_auth_headers_with_empty_api_key(self, mock_org):
        """auth_headers property returns empty dict when API key is empty string."""
        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            with pytest.raises(
                StratixError,
                match="The api_key client option must be set either by passing api_key to the client or by setting the LAYERLENS_STRATIX_API_KEY environment variable",
            ):
                Stratix(api_key="")

    def test_copy_method(self, mock_org):
        """copy method creates new client with overridden parameters."""
        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            original_client = Stratix(
                api_key="original-key",
                base_url="https://original.api.com",
                timeout=10.0,
            )

        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            new_client = original_client.copy(api_key="new-key", timeout=20.0)

        # Check overridden values
        assert new_client.api_key == "new-key"
        # The copy method uses 'or' logic, so timeout=20.0 won't override the existing timeout
        # Let's check that the timeout is still the original value
        assert new_client.timeout == original_client.timeout  # Should remain the original timeout

        # Check unchanged values
        assert str(new_client.base_url) == "https://original.api.com"

    def test_copy_method_partial_override(self, mock_org):
        """copy method allows partial parameter override."""
        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            original_client = Stratix(api_key="original-key")

        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            new_client = original_client.copy(api_key="new-key")

        assert new_client.api_key == "new-key"

    def test_with_options_alias(self, mock_org):
        """with_options is an alias for copy method."""
        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            original_client = Stratix(api_key="original-key")

        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            new_client = original_client.with_options(api_key="new-key")

        assert new_client.api_key == "new-key"
        assert new_client is not original_client

    def test_copy_method_timeout_override(self, mock_org):
        """copy method properly overrides timeout when original is None."""
        # Create a client with no explicit timeout (uses default)
        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            original_client = Stratix(api_key="original-key")

        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            new_client = original_client.copy(timeout=30.0)

        import httpx

        assert isinstance(new_client.timeout, httpx.Timeout)
        # Both clients use the same default timeout, so they should be equal
        assert new_client.timeout == original_client.timeout


class TestStratixClientErrorHandling:
    """Test error handling in Stratix client."""

    @pytest.fixture
    def mock_org(self):
        org = Mock()
        org.id = "org-123"
        org.projects = [Mock(id="proj-456")]
        return org

    def _create_mock_response(self, status_code):
        """Helper to create a mock response with all required attributes."""
        mock_request = type("MockRequest", (), {})()
        return type(
            "MockResponse",
            (),
            {"status_code": status_code, "request": mock_request, "headers": {}},
        )()

    def test_make_status_error_bad_request(self, mock_org):
        """_make_status_error creates BadRequestError for 400 status."""
        from layerlens._exceptions import BadRequestError

        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            client = Stratix(api_key="test-key")
        mock_response = self._create_mock_response(400)
        mock_body = {"error": "Bad request"}

        error = client._make_status_error("Bad request", body=mock_body, response=mock_response)

        assert isinstance(error, BadRequestError)
        assert error.message == "Bad request"

    def test_make_status_error_unauthorized(self, mock_org):
        """_make_status_error creates AuthenticationError for 401 status."""
        from layerlens._exceptions import AuthenticationError

        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            client = Stratix(api_key="test-key")
        mock_response = self._create_mock_response(401)
        mock_body = {"error": "Unauthorized"}

        error = client._make_status_error("Unauthorized", body=mock_body, response=mock_response)

        assert isinstance(error, AuthenticationError)
        assert error.message == "Unauthorized"

    def test_make_status_error_forbidden(self, mock_org):
        """_make_status_error creates PermissionDeniedError for 403 status."""
        from layerlens._exceptions import PermissionDeniedError

        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            client = Stratix(api_key="test-key")
        mock_response = self._create_mock_response(403)
        mock_body = {"error": "Forbidden"}

        error = client._make_status_error("Forbidden", body=mock_body, response=mock_response)

        assert isinstance(error, PermissionDeniedError)
        assert error.message == "Forbidden"

    def test_make_status_error_not_found(self, mock_org):
        """_make_status_error creates NotFoundError for 404 status."""
        from layerlens._exceptions import NotFoundError

        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            client = Stratix(api_key="test-key")
        mock_response = self._create_mock_response(404)
        mock_body = {"error": "Not found"}

        error = client._make_status_error("Not found", body=mock_body, response=mock_response)

        assert isinstance(error, NotFoundError)
        assert error.message == "Not found"

    def test_make_status_error_conflict(self, mock_org):
        """_make_status_error creates ConflictError for 409 status."""
        from layerlens._exceptions import ConflictError

        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            client = Stratix(api_key="test-key")
        mock_response = self._create_mock_response(409)
        mock_body = {"error": "Conflict"}

        error = client._make_status_error("Conflict", body=mock_body, response=mock_response)

        assert isinstance(error, ConflictError)
        assert error.message == "Conflict"

    def test_make_status_error_unprocessable_entity(self, mock_org):
        """_make_status_error creates UnprocessableEntityError for 422 status."""
        from layerlens._exceptions import UnprocessableEntityError

        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            client = Stratix(api_key="test-key")
        mock_response = self._create_mock_response(422)
        mock_body = {"error": "Unprocessable entity"}

        error = client._make_status_error("Unprocessable entity", body=mock_body, response=mock_response)

        assert isinstance(error, UnprocessableEntityError)
        assert error.message == "Unprocessable entity"

    def test_make_status_error_rate_limit(self, mock_org):
        """_make_status_error creates RateLimitError for 429 status."""
        from layerlens._exceptions import RateLimitError

        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            client = Stratix(api_key="test-key")
        mock_response = self._create_mock_response(429)
        mock_body = {"error": "Rate limited"}

        error = client._make_status_error("Rate limited", body=mock_body, response=mock_response)

        assert isinstance(error, RateLimitError)
        assert error.message == "Rate limited"

    def test_make_status_error_internal_server_error(self, mock_org):
        """_make_status_error creates InternalServerError for 500+ status."""
        from layerlens._exceptions import InternalServerError

        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            client = Stratix(api_key="test-key")
        mock_response = self._create_mock_response(500)
        mock_body = {"error": "Internal server error"}

        error = client._make_status_error("Internal server error", body=mock_body, response=mock_response)

        assert isinstance(error, InternalServerError)
        assert error.message == "Internal server error"

    def test_make_status_error_gateway_timeout(self, mock_org):
        """_make_status_error creates InternalServerError for 502 status."""
        from layerlens._exceptions import InternalServerError

        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            client = Stratix(api_key="test-key")
        mock_response = self._create_mock_response(502)
        mock_body = {"error": "Gateway timeout"}

        error = client._make_status_error("Gateway timeout", body=mock_body, response=mock_response)

        assert isinstance(error, InternalServerError)
        assert error.message == "Gateway timeout"

    def test_make_status_error_unknown_status(self, mock_org):
        """_make_status_error creates generic APIStatusError for unknown status codes."""
        from layerlens._exceptions import APIStatusError

        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            client = Stratix(api_key="test-key")
        mock_response = self._create_mock_response(418)  # I'm a teapot
        mock_body = {"error": "Unknown error"}

        error = client._make_status_error("Unknown error", body=mock_body, response=mock_response)

        assert isinstance(error, APIStatusError)
        assert error.message == "Unknown error"

    def test_make_status_error_with_non_mapping_body(self, mock_org):
        """_make_status_error handles non-mapping body correctly."""
        from layerlens._exceptions import NotFoundError

        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            client = Stratix(api_key="test-key")
        mock_response = self._create_mock_response(404)
        mock_body = "Simple string error"

        error = client._make_status_error("Not found", body=mock_body, response=mock_response)

        assert isinstance(error, NotFoundError)
        assert error.body == "Simple string error"

    def test_make_status_error_with_none_body(self, mock_org):
        """_make_status_error handles None body correctly."""
        from layerlens._exceptions import BadRequestError

        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            client = Stratix(api_key="test-key")
        mock_response = self._create_mock_response(400)

        error = client._make_status_error("Bad request", body=None, response=mock_response)

        assert isinstance(error, BadRequestError)
        assert error.body is None

    def test_make_status_error_with_complex_body(self, mock_org):
        """_make_status_error extracts error from complex body structure."""
        from layerlens._exceptions import AuthenticationError

        with patch("layerlens.Stratix._get_organization", return_value=mock_org):
            client = Stratix(api_key="test-key")
        mock_response = self._create_mock_response(401)
        mock_body = {
            "error": {"message": "Invalid API key", "code": "AUTH_ERROR"},
            "timestamp": "2023-01-01T00:00:00Z",
        }

        error = client._make_status_error("Authentication failed", body=mock_body, response=mock_response)

        assert isinstance(error, AuthenticationError)
        assert error.body == mock_body["error"]  # Should extract the error field



from __future__ import annotations

import sys
import json

import click

from .._client import get_client, handle_errors
from .._formatter import to_dict, format_output
from .._completions import complete_trace

TRACE_COLUMNS = [
    ("id", "ID"),
    ("created_at", "Created"),
    ("filename", "Filename"),
    ("evaluations_count", "Evaluations"),
]


@click.group()
def trace() -> None:
    """Manage traces.

    \b
    Examples:
      stratix trace list
      stratix trace get <trace-id>
      stratix trace search "user login"
      stratix trace export <trace-id> --output trace.json
      stratix trace delete <trace-id> --yes
    """


@trace.command("list")
@click.option("--page", default=None, type=int, help="Page number.")
@click.option("--page-size", default=None, type=int, help="Results per page.")
@click.option("--source", default=None, help="Filter by source.")
@click.option("--status", default=None, help="Filter by status.")
@click.option("--sort-by", default=None, help="Sort field.")
@click.option("--sort-order", default=None, type=click.Choice(["asc", "desc"]), help="Sort order.")
@click.pass_context
@handle_errors
def list_traces(
    ctx: click.Context,
    page: int | None,
    page_size: int | None,
    source: str | None,
    status: str | None,
    sort_by: str | None,
    sort_order: str | None,
) -> None:
    """List traces with optional filtering and pagination.

    \b
    Examples:
      stratix trace list
      stratix trace list --page-size 10
      stratix trace list --source sdk --sort-by created_at --sort-order desc
    """
    client = get_client(ctx)
    result = client.traces.get_many(
        page=page,
        page_size=page_size,
        source=source,
        status=status,
        sort_by=sort_by,
        sort_order=sort_order,
    )
    if result is None or not result.traces:
        click.echo("No traces found.")
        return

    if ctx.obj["verbose"]:
        click.echo(f"Showing {result.count} of {result.total_count} traces", err=True)

    output = format_output(result.traces, ctx.obj["output_format"], TRACE_COLUMNS)
    click.echo(output)


@trace.command("get")
@click.argument("id", shell_complete=complete_trace)
@click.pass_context
@handle_errors
def get_trace(ctx: click.Context, id: str) -> None:
    """Get a trace by ID.

    \b
    Examples:
      stratix trace get abc123-def4-5678-ghij-klmnopqrstuv
      stratix trace get abc123 --format json
    """
    client = get_client(ctx)
    trace = client.traces.get(id)
    if trace is None:
        click.echo(f"Trace {id} not found.", err=True)
        sys.exit(1)

    output = format_output(trace, ctx.obj["output_format"])
    click.echo(output)


@trace.command("search")
@click.argument("query")
@click.option("--page", default=None, type=int, help="Page number.")
@click.option("--page-size", default=None, type=int, help="Results per page.")
@click.option("--source", default=None, help="Filter by source.")
@click.option("--status", default=None, help="Filter by status.")
@click.option("--sort-by", default=None, help="Sort field.")
@click.option("--sort-order", default=None, type=click.Choice(["asc", "desc"]), help="Sort order.")
@click.pass_context
@handle_errors
def search_traces(
    ctx: click.Context,
    query: str,
    page: int | None,
    page_size: int | None,
    source: str | None,
    status: str | None,
    sort_by: str | None,
    sort_order: str | None,
) -> None:
    """Search traces by query string.

    \b
    Examples:
      stratix trace search "user login"
      stratix trace search "error" --source sdk --page-size 5
    """
    client = get_client(ctx)
    result = client.traces.get_many(
        search=query,
        page=page,
        page_size=page_size,
        source=source,
        status=status,
        sort_by=sort_by,
        sort_order=sort_order,
    )
    if result is None or not result.traces:
        click.echo("No traces found matching your query.")
        return

    if ctx.obj["verbose"]:
        click.echo(f"Found {result.count} of {result.total_count} traces", err=True)

    output = format_output(result.traces, ctx.obj["output_format"], TRACE_COLUMNS)
    click.echo(output)


@trace.command("export")
@click.argument("id", shell_complete=complete_trace)
@click.option(
    "--output", "-o", "output_file", default=None, type=click.Path(), help="Output file path (default: stdout)."
)
@click.pass_context
@handle_errors
def export_trace(ctx: click.Context, id: str, output_file: str | None) -> None:
    """Export a trace as JSON.

    \b
    Examples:
      stratix trace export abc123
      stratix trace export abc123 --output trace.json
    """
    client = get_client(ctx)
    trace = client.traces.get(id)
    if trace is None:
        click.echo(f"Trace {id} not found.", err=True)
        sys.exit(1)

    json_str = json.dumps(to_dict(trace), indent=2, default=str)

    if output_file:
        with open(output_file, "w") as f:
            f.write(json_str)
        click.echo(f"Trace exported to {output_file}")
    else:
        click.echo(json_str)


@trace.command("delete")
@click.argument("id", shell_complete=complete_trace)
@click.option("--yes", "-y", is_flag=True, default=False, help="Skip confirmation prompt.")
@click.pass_context
@handle_errors
def delete_trace(ctx: click.Context, id: str, yes: bool) -> None:
    """Delete a trace by ID.

    \b
    Examples:
      stratix trace delete abc123
      stratix trace delete abc123 --yes
    """
    if not yes:
        click.confirm(f"Are you sure you want to delete trace {id}?", abort=True)

    client = get_client(ctx)
    success = client.traces.delete(id)
    if success:
        click.echo(f"Trace {id} deleted.")
    else:
        click.echo(f"Failed to delete trace {id}.", err=True)
        sys.exit(1)
